from hazma.decay_helper_functions.decay_charged_pion \
    cimport CSpectrum as cp_spec
from hazma.decay_helper_functions.decay_neutral_pion \
    cimport CSpectrum as np_spec
from hazma.decay_helper_functions.decay_neutral_pion cimport CSpectrumPoint
from hazma.decay_helper_functions.decay_muon cimport CSpectrum as mu_spec

import cython
import numpy as np
cimport numpy as np
from scipy.integrate import quad

from libc.math cimport exp, log, M_PI, log10, sqrt, abs, pow

include "../decay_helper_functions/parameters.pxd"

cdef double mmu = MASS_MU
cdef double me = MASS_E
cdef double mpi = MASS_PI
cdef double qe = sqrt(4. * M_PI * ALPHA_EM)
ctypedef np.ndarray ndarray

cdef int n_interp_pts = 500

cdef np.ndarray __e_gams = np.zeros((n_interp_pts,), dtype=np.float64)

cdef np.ndarray __spec_cp = np.zeros((n_interp_pts,), dtype=np.float64)
cdef np.ndarray __spec_np = np.zeros((n_interp_pts,), dtype=np.float64)
cdef np.ndarray __spec_mu = np.zeros((n_interp_pts,), dtype=np.float64)

cdef double __set_spectra(double ms):
    global __e_gams
    global __spec_cp
    global __spec_np
    global __spec_mu

    __e_gams = np.logspace(-1., np.log10(ms / 2.), num=n_interp_pts)

    __spec_cp = cp_spec(__e_gams, ms / 2., "total")
    __spec_np = np_spec(__e_gams, ms / 2.)
    __spec_mu = mu_spec(__e_gams, ms / 2.)

cdef double __interp_spec(double eng_gam, str fs):
    """
    Intepolation function for the short kaon data.

    Parameters
    ----------
    eng_gam : double
        Energy of the photon.
    fs : str {"total"}
        String specifying which decay fs to use.
    """
    if fs == "cp":
        if eng_gam < 10**-1:
            return __spec_cp[0] * __e_gams[0] / eng_gam
        return np.interp(eng_gam, __e_gams, __spec_cp)
    if fs == "mu":
        if eng_gam < 10**-1:
            return __spec_mu[0] * __e_gams[0] / eng_gam
        return np.interp(eng_gam, __e_gams, __spec_mu)
    else:
        return 0.0

@cython.cdivision(True)
cdef double __dnde_fsr_cp_srf(double egam, double ms):
    cdef double mupi = mpi / ms
    cdef double x = 2. * egam / ms
    cdef double xmin = 0.0
    cdef double xmax = 1 - 4. * mupi**2
    cdef double dynamic
    cdef double coeff
    cdef double result

    if x < xmin or x > xmax:
        return 0.0

    dynamic = (
                      -2 * sqrt(1 - x) * sqrt(1 - 4 * mupi**2 - x) +
                      (-1 + 2 * mupi**2 + x) * log(
                  (1 - x - sqrt(1 - x) * sqrt(1 - 4 * mupi**2 - x))**2 /
                  (-1 + x - sqrt(1 - x) * sqrt(1 - 4 * mupi**2 - x))**2)) / x

    coeff = qe**2 / (8. * sqrt(1 - 4 * mupi**2) * M_PI**2)

    result = dynamic * coeff

    return 2 * result / ms

@cython.cdivision(True)
cdef double __dnde_fsr_l_srf(double egam, double ml, double ms):
    cdef double mul = ml / ms
    cdef double x = 2. * egam / ms

    cdef double xmin = 0.0
    cdef double xmax = 1 - 4. * mul**2

    cdef double dynamic
    cdef double coeff
    cdef double result

    if x < xmin or x > xmax:
        return 0.0

    dynamic = (
                      4 * (-1 + 4 * mul**2) *
                      sqrt(1 - x) * sqrt(1 - 4 * mul**2 - x) +
                      (2 - 12 * mul**2 + 16 * mul**4 - 2 * x + 8 * mul**2 * x + x**2) * log(
                  (1 - x + sqrt(
                      (-1 + x) * (-1 + 4 * mul**2 + x)))**2 / (-1 + x + sqrt(
                      (-1 + x) * (-1 + 4 * mul**2 + x)))**2)) / x

    coeff = qe**2 / (16. * (1 - 4 * mul**2)**1.5 * M_PI**2)

    result = dynamic * coeff

    return 2 * result / ms

@cython.cdivision(True)
@cython.boundscheck(True)
@cython.wraparound(False)
cdef double __integrand(double cl, double eng_gam, double eng_s,
                        double ms, np.ndarray[double] pws, str fs):
    """
    Integrand of the boost integralself.

    Parameters
    ----------
    cl : float
        Angle the final state particle make with respect to the z-axis.
    eng_gams : float
        Gamma-ray energy to evaluate spectrum at.
    eng_s : float
        Energy of the scalar mediator.
    params :
        Scalar mediator model parameters.

    Returns
    -------
    integrand : float
        The value of the boost integral.
    """
    cdef double pwee = pws[0]
    cdef double pwmumu = pws[1]
    cdef double pwpi0pi0 = pws[2]
    cdef double pwpipi = pws[3]

    cdef double beta = sqrt(1. - (ms / eng_s)**2)
    cdef double gamma = eng_s / ms

    cdef double jac = 1. / (2. * gamma * abs(1. - beta * cl))

    cdef double eng_gam_srf = eng_gam * gamma * (1. - beta * cl)

    cdef double dnde = 0.0

    cdef double dnde_ee_f = pwee * __dnde_fsr_l_srf(eng_gam_srf, me, ms)
    cdef double dnde_mu_f = pwmumu * __dnde_fsr_l_srf(eng_gam_srf, mmu, ms)

    cdef double dnde_cp_f = pwpipi * __dnde_fsr_cp_srf(eng_gam_srf, ms)

    cdef double dnde_cp_d = 2. * pwpipi * __interp_spec(eng_gam_srf, "cp")
    cdef double dnde_np_d = 2. * pwpi0pi0 * CSpectrumPoint(eng_gam_srf, ms / 2.)
    cdef double dnde_mu_d = 2. * pwmumu * __interp_spec(eng_gam_srf, "mu")

    dnde = dnde_ee_f + dnde_mu_f + dnde_cp_f + \
           dnde_cp_d + dnde_np_d + dnde_mu_d

    if fs == "total":
        return jac * dnde
    if fs == "e e g":
        return jac * dnde_ee_f
    if fs == "pi pi g":
        return jac * dnde_cp_f
    if fs == "pi pi":
        return jac * dnde_cp_d
    if fs == "pi0 pi0":
        return jac * dnde_np_d
    if fs == "mu mu g":
        return jac * dnde_mu_f
    if fs == "mu mu":
        return jac * dnde_mu_d
    if fs == "g g":
        return 0.0

@cython.boundscheck(True)
@cython.wraparound(False)
@cython.cdivision(True)
cdef double __dnde_decay_s(double eng_gam, double eng_s, double ms,
                           np.ndarray[double] pws, str fs):
    """
    Unvectorized dnde_decay_s

    Compute the gamma ray spectrum from the decay of the scalar mediator.

    Parameters
    ----------
    eng_gams : float or array-like
        Float or array of gamma-ray energies to evaluate spectrum at.
    eng_s : float
        Energy of the scalar mediator.
    params :
        Scalar mediator model parameters.

    Returns
    -------
    dnde : float or array-like
        Values of dnde at gamma-ray energies `eng_gams`.
    """
    cdef double lines_contrib = 0.0
    cdef double beta = sqrt(1. - pow(ms / eng_s, 2.))
    cdef double gamma = eng_s / ms

    if eng_s < ms:
        return 0.

    cdef double eplus = eng_s * (1. + beta) / 2.0
    cdef double eminus = eng_s * (1. - beta) / 2.0
    cdef double result = 0.0

    if eminus <= eng_gam <= eplus:
        lines_contrib = pws[4] * 1. / (eng_s * beta)

    if fs == "g g":
        return lines_contrib

    if fs == "total":
        result = quad(__integrand, -1.0, 1.0, points=[-1.0, 1.0],
                      args=(eng_gam, eng_s, ms, pws, fs), epsabs=10**-10.,
                      epsrel=10**-5.)[0]

        return result + lines_contrib

    return result

@cython.boundscheck(True)
@cython.wraparound(False)
def dnde_decay_s_pt(double eng_gam, double eng_s, double ms,
                    np.ndarray[double] pws, str fs):
    """
    Compute the gamma ray spectrum from the decay of the scalar mediator.

    Parameters
    ----------
    eng_gams : float
        Gamma-ray energy to evaluate spectrum at.
    eng_s : float
        Energy of the scalar mediator.
    params :
        Scalar mediator model parameters.

    Returns
    -------
    dnde : float or array-like
        Value of dnde at gamma-ray energy `eng_gam`.
    """
    __set_spectra(ms)
    return __dnde_decay_s(eng_gam, eng_s, ms, pws, fs)

@cython.boundscheck(True)
@cython.wraparound(False)
def dnde_decay_s(np.ndarray[double] eng_gam, double eng_s, double ms,
                 np.ndarray[double] pws, str fs):
    """
    Compute the gamma ray spectrum from the decay of the scalar mediator.

    Parameters
    ----------
    eng_gams : float
        Gamma-ray energy to evaluate spectrum at.
    eng_s : float
        Energy of the scalar mediator.
    params :
        Scalar mediator model parameters.

    Returns
    -------
    dnde : float or array-like
        Value of dnde at gamma-ray energy `eng_gam`.
    """
    __set_spectra(ms)
    cdef int num_pts = len(eng_gam)
    cdef int i

    spec = np.zeros(num_pts, dtype=np.float64)

    for i in range(num_pts):
        spec[i] = __dnde_decay_s(eng_gam[i], eng_s, ms, pws, fs)

    return spec



import numpy as np
cimport numpy as np

cdef double __spectrum_rf(double)
cdef double __integrand(double, double, double)
cdef double CSpectrumPoint(double, double)
cdef np.ndarray CSpectrum(np.ndarray, double)



from hazma.theory import Theory

import warnings
from hazma.hazma_errors import PreAlphaWarning

from hazma.axial_vector_mediator._axial_vector_mediator_cross_sections import (
    AxialVectorMediatorCrossSections,
)
from hazma.axial_vector_mediator._axial_vector_mediator_positron_spectra import (
    AxialVectorMediatorPositronSpectra,
)
from hazma.axial_vector_mediator._axial_vector_mediator_spectra import (
    AxialVectorMediatorSpectra,
)
from hazma.axial_vector_mediator._axial_vector_mediator_widths import (
    AxialVectorMediatorWidths,
)


class AxialVectorMediator(
    AxialVectorMediatorCrossSections,
    AxialVectorMediatorPositronSpectra,
    AxialVectorMediatorSpectra,
    AxialVectorMediatorWidths,
    Theory,
):
    r"""
    Create an axial vector mediator model object.
    """

    def __init__(self, mx, ma, gaxx, gauu, gadd, gass, gaee, gamumu):
        self._mx = mx
        self._ma = ma
        self._gaxx = gaxx
        self._gauu = gauu
        self._gadd = gadd
        self._gass = gass
        self._gaee = gaee
        self._gamumu = gamumu
        self.compute_width_a()

    @property
    def mx(self):
        return self._mx

    @mx.setter
    def mx(self, mx):
        self._mx = mx
        self.compute_width_a()

    @property
    def ma(self):
        return self._ma

    @ma.setter
    def ma(self, ma):
        self._ma = ma
        self.compute_width_a()

    @property
    def gaxx(self):
        return self._gaxx

    @gaxx.setter
    def gaxx(self, gaxx):
        self._gaxx = gaxx
        self.compute_width_a()

    @property
    def gauu(self):
        return self._gauu

    @gauu.setter
    def gauu(self, gauu):
        self._gauu = gauu
        self.compute_width_a()

    @property
    def gadd(self):
        return self._gadd

    @gadd.setter
    def gadd(self, gadd):
        self._gadd = gadd
        self.compute_width_a()

    @property
    def gass(self):
        return self._gass

    @gass.setter
    def gass(self, gass):
        self._gass = gass
        self.compute_width_a()

    @property
    def gaee(self):
        return self._gaee

    @gaee.setter
    def gaee(self, gaee):
        self._gaee = gaee
        self.compute_width_a()

    @property
    def gamumu(self):
        return self._gamumu

    @gamumu.setter
    def gamumu(self, gamumu):
        self._gamumu = gamumu
        self.compute_width_a()

    def description(self):
        warnings.warn("", PreAlphaWarning)
        pass

    @classmethod
    def list_annihilation_final_states(cls):
        """
        Return a list of the available final states.

        Returns
        -------
        fs : array-like
            Array of the available final states.
        """
        return ["pi0 pi pi"]

    def constraints(self):
        pass

    def constrain(self, p1, p1_vals, p2, p2_vals, ls_or_img="image"):
        pass



from cmath import sqrt, pi

from hazma.parameters import vh, b0, alpha_em
from hazma.parameters import charged_pion_mass as mpi
from hazma.parameters import neutral_pion_mass as mpi0
from hazma.parameters import up_quark_mass as muq
from hazma.parameters import down_quark_mass as mdq
from hazma.parameters import electron_mass as me
from hazma.parameters import muon_mass as mmu


class ScalarMediatorWidths:
    def width_s_to_gg(self):
        """
        Returns the partial decay width of the scalar decaying into photon.
        """
        ret_val = (alpha_em ** 2 * self.gsFF ** 2 * (self.ms ** 2) ** 1.5) / (
            128.0 * self.lam ** 2 * pi ** 3
        )

        assert ret_val.imag == 0
        assert ret_val.real >= 0

        return ret_val.real

    def width_s_to_pi0pi0(self):
        """
        Returns the partial decay width of the scalar decaying into
        neutral pions.
        """
        ms = self.ms

        if ms > 2.0 * mpi0:
            gsff = self.gsff
            gsGG = self.gsGG
            vs = self.vs
            Lam = self.lam

            ret_val = (
                sqrt(-4 * mpi0 ** 2 + ms ** 2)
                * (
                    (gsGG * (4 * mpi0 ** 2 - 2 * ms ** 2)) / (9 * Lam + 4 * gsGG * vs)
                    + (
                        b0
                        * (mdq + muq)
                        * (
                            27 * gsff ** 2 * Lam ** 2 * vs * (3 * Lam + 4 * gsGG * vs)
                            - 2
                            * gsGG
                            * vh ** 2
                            * (
                                27 * Lam ** 2
                                - 30 * gsGG * Lam * vs
                                + 8 * gsGG ** 2 * vs ** 2
                            )
                            + gsff
                            * (
                                -81 * Lam ** 3 * vh
                                + 48 * gsGG ** 2 * Lam * vh * vs ** 2
                            )
                        )
                    )
                    / (81.0 * Lam ** 3 * vh ** 2)
                )
                ** 2
            ) / (16.0 * ms ** 2 * pi)

            assert ret_val.imag == 0
            assert ret_val.real >= 0

            return ret_val.real
        else:
            return 0.0

    def width_s_to_pipi(self):
        """
        Returns the partial decay width of the scalar decaying into
        charged pion.
        """
        ms = self.ms

        if ms > 2.0 * mpi:
            gsff = self.gsff
            gsGG = self.gsGG
            vs = self.vs
            Lam = self.lam

            ret_val = (
                sqrt(-4 * mpi ** 2 + ms ** 2)
                * (
                    (gsGG * (4 * mpi ** 2 - 2 * ms ** 2)) / (9 * Lam + 4 * gsGG * vs)
                    + (
                        b0
                        * (mdq + muq)
                        * (
                            27 * gsff ** 2 * Lam ** 2 * vs * (3 * Lam + 4 * gsGG * vs)
                            - 2
                            * gsGG
                            * vh ** 2
                            * (
                                27 * Lam ** 2
                                - 30 * gsGG * Lam * vs
                                + 8 * gsGG ** 2 * vs ** 2
                            )
                            + gsff
                            * (
                                -81 * Lam ** 3 * vh
                                + 48 * gsGG ** 2 * Lam * vh * vs ** 2
                            )
                        )
                    )
                    / (81.0 * Lam ** 3 * vh ** 2)
                )
                ** 2
            ) / (16.0 * ms ** 2 * pi)

            assert ret_val.imag == 0
            assert ret_val.real >= 0

            return ret_val.real
        else:
            return 0.0

    def width_s_to_xx(self):
        """
        Returns the partial decay width of the scalar decaying into
        two fermions x.
        """

        ms = self.ms
        mx = self.mx

        if ms > 2.0 * mx:
            gsxx = self.gsxx

            ret_val = (
                gsxx ** 2 * (ms - 2 * mx) * (ms + 2 * mx) * sqrt(ms ** 2 - 4 * mx ** 2)
            ) / (32.0 * ms ** 2 * pi)

            assert ret_val.imag == 0
            assert ret_val.real >= 0

            return ret_val.real
        else:
            return 0.0

    def width_s_to_ff(self, f):
        """
        Returns the partial decay width of the scalar decaying into
        two fermions x.

        Parameters
        ----------
        f : string
            Name of the final state fermion.
        """
        ms = self.ms

        if f == "e":
            mf = me
        elif f == "mu":
            mf = mmu

        if ms > 2.0 * mf:
            gsff = self.gsff

            ret_val = -(
                gsff ** 2
                * mf ** 2
                * (2 * mf - ms)
                * (2 * mf + ms)
                * sqrt(-4 * mf ** 2 + ms ** 2)
            ) / (32.0 * ms ** 2 * pi * vh ** 2)

            assert ret_val.imag == 0
            assert ret_val.real >= 0

            return ret_val.real
        else:
            return 0.0

    def partial_widths(self):
        """
        Returns a dictionary for the partial decay widths of the scalar
        mediator.

        Returns
        -------
        width_dict : dictionary
            Dictionary of all of the individual decay widths of the scalar
            mediator as well as the total decay width. The possible decay
            modes of the scalar mediator are 'g g', 'pi0 pi0', 'pi pi', 'x x'
            and 'f f'. The total decay width has the key 'total'.
        """
        w_gg = self.width_s_to_gg()
        w_pi0pi0 = self.width_s_to_pi0pi0()
        w_pipi = self.width_s_to_pipi()
        w_xx = self.width_s_to_xx()

        w_ee = self.width_s_to_ff("e")
        w_mumu = self.width_s_to_ff("mu")

        total = w_gg + w_pi0pi0 + w_pipi + w_xx + w_ee + w_mumu

        width_dict = {
            "g g": w_gg,
            "pi0 pi0": w_pi0pi0,
            "pi pi": w_pipi,
            "x x": w_xx,
            "e e": w_ee,
            "mu mu": w_mumu,
            "total": total,
        }

        return width_dict



"""Module for computing gamma ray spectra from a many-particle final state.

"""
# author : Logan Morrison and Adam Coogan
# date : December 2017

from hazma import rambo
from hazma.gamma_ray_helper_functions.gamma_ray_generator import gamma, gamma_point
from hazma.field_theory_helper_functions.common_functions import cross_section_prefactor
import numpy as np


def gamma_ray_decay(
    particles,
    cme,
    photon_energies,
    mat_elem_sqrd=lambda k_list: 1.0,
    num_ps_pts=1000,
    num_bins=25,
    verbose=False,
):
    r"""Returns gamma ray spectrum from the decay of a set of particles.

    This function works by running the Monte-Carlo algorithm RAMBO on the
    final state particles to obtain energies distributions for each final state
    particle. The energies distributions are then convolved with the decay
    spectra in order to obtain the gamma-ray spectrum.

    Parameters
    ----------
    particles : `array_like`
        List of particle names. Available particles are 'muon', 'electron'
        'charged_pion', 'neutral pion', 'charged_kaon', 'long_kaon', 'short_kaon'.
    cme : double
        Center of mass energy of the final state in MeV.
    photon_energies : np.ndarray[double, ndim=1]
        List of photon energies in MeV to evaluate spectra at.
    mat_elem_sqrd : double(\*func)(np.ndarray, )
        Function for the matrix element squared of the process. Must be a
        function taking in a list of four momenta of shape (len(particles), 4).
        Default value is a flat matrix element with a value of 1.
    num_ps_pts : int {1000}, optional
        Number of phase space points to use.
    num_bins : int {25}, optional
        Number of bins to use.
    verbose: Bool
        If true, additional output is displayed while the function is computing
        spectra.

    Returns
    -------
    spec : np.ndarray
        Total gamma ray spectrum from all final state particles.

    Notes
    -----
    The total spectrum is computed using

    .. math::

        \frac{dN}{dE}(E_{\gamma})
        =\sum_{i,j}P_{i}(E_{j})\frac{dN_i}{dE}(E_{\gamma}, E_{j})

    where :math:`i` runs over the final state particles, :math:`j` runs over
    energies sampled from probability distributions. :math:`P_{i}(E_{j})` is
    the probability that particle :math:`i` has energy :math:`E_{j}`. The
    probabilities are computed using ``hazma.phase_space_generator.rambo``. The
    total number of energies used is ``num_bins``.

    Examples
    --------

    Example of generating a spectrum from a muon, charged kaon and long kaon
    with total energy of 5000 MeV::

        from hazma.gamma_ray import gamma_ray_decay
        import numpy as np
        particles = np.array(['muon', 'charged_kaon', 'long_kaon'])
        cme = 5000.
        photon_energies = np.logspace(0., np.log10(cme), num=200, dtype=np.float64)
        gamma_ray_decay(particles, cme, photon_energies)

    """
    if type(particles) == str:
        particles = [particles]

    particles = np.array(particles)

    if hasattr(photon_energies, "__len__"):
        return gamma(
            particles,
            cme,
            photon_energies,
            mat_elem_sqrd=mat_elem_sqrd,
            num_ps_pts=num_ps_pts,
            num_bins=num_bins,
            verbose=verbose,
        )
    return gamma_point(
        particles, cme, photon_energies, mat_elem_sqrd, num_ps_pts, num_bins
    )


def gamma_ray_fsr(
    isp_masses,
    fsp_masses,
    cme,
    mat_elem_sqrd_tree=lambda k_list: 1.0,
    mat_elem_sqrd_rad=lambda k_list: 1.0,
    num_ps_pts=1000,
    num_bins=25,
):
    r"""Returns the FSR spectrum for a user-specified particle physics process.

    This is done by first running a Monte Carlo to compute the non-radiative
    cross cross_section, :math:`\sigma(I \to F)` from the non-radiative squared
    matrix element: :math:`|M(I \to F)|^2`. The differential radiative cross
    section, :math:`\frac{d\sigma(I \to F + \gamma)}{dE_{\gamma}}`, is then
    computed using a Monte Carlo using the radiative squared matrix element
    :math:`|M(I \to F + \gamma)|^2`. The spectrum is then

    .. math:: \frac{dN}{dE_{\gamma}} = \frac{1}{\sigma(I \to F)}
        \frac{d\sigma(I \to F + \gamma)}{dE_{\gamma}}

    Parameters
    ----------
    isp_masses : np.ndarray[double, ndim=1]
        Array of masses of the initial state particles.
    fsp_masses : np.ndarray[double, ndim=1]
        Array of masses of the final state particles. Note that the photon
        should be the last mass.
    cme : double
        Center of mass energy of the process.
    mat_elem_sqrd_tree : double(*)(np.ndarray[double, ndim=1])
        Tree level squared matrix element.
    mat_elem_sqrd_rad : double(*)(np.ndarray[double, ndim=1])
        Radiative squared matrix element.
    num_ps_pts : int
        Number of Monte Carlo events to generate.
    num_bins : int
        Number of gamma ray energies to use.

    Returns
    -------
    photon_energies : np.ndarray[double, ndim=1]
        Array of the gamma ray energies.
    spectrum : np.ndarray[double, ndim=1]
        Array of the spectrum values evaluated at the gamma ray energies.

    Examples
    --------

    Compute spectrum from dark matter annihilating into a pair of charged
    pion through via an effective operator:

        .. math:: \frac{c_1}{\Lambda}\chi\bar{\chi}\pi^{+}\pi^{-}

    Step up a class to compute the tree and radiative matrix elements. We could
    just specify the functions, but we would have to use globals to specify
    parameters. We find a class easier::


        from hazma.gamma_ray import gamma_ray_fsr
        from hazma.parameters import charged_pion_mass as mpi
        from hazma.parameters import qe # electric charge
        from hazma.field_theory_helper_functions.common_functions \
            import minkowski_dot as MDot
        class Msqrds:
            def __init__(self, mx, c1, lam):
                self.mx = mx
                self.c1 = c1
                self.lam = lam
            def tree(self, momenta):
                ppi1 = momenta[0] # first charged pion four-momentum
                ppi2 = momenta[1] # second charged pion four-momentum
                cme = ppi1[0] + ppi2[0] # center-of-mass energy
                return c1**2 * (cme**2 - 4 * self.mx**2) / (2 * self.lam**2)
            def radiative(self, momenta):
                ppi1 = momenta[0] # first charged pion four-momentum
                ppi2 = momenta[1] # second charged pion four-momentum
                k = momenta[2] # photon four-momentum
                Q = ppi1[0] + ppi2[0] + k[0] # center-of-mass energy
                mux = self.mx / Q
                mupi = mpi / Q
                s = MDot(ppi1 + ppi2, ppi1 + ppi2)
                t = MDot(ppi1 + k, ppi1 + k)
                u = MDot(ppi2 + k, ppi2 + k)
                return ((2 * self.c1**2 * (-1 + 4*mux**2) * Q**2 * qe**2 *
                        (s * (-(mupi**2 * Q**2) + t) * (mupi**2 * Q**2 - u) +
                        (-2 * mupi**3 * Q**3 + mupi * Q * (t + u))**2)) /
                        (self.lam**2 * (-(mupi**2 * Q**2) + t)**2 *
                        (-(mupi**2 * Q**2) + u)**2)

    Now we instantiate an object for the matrix elements, specify the
    parameters to pass to ``gamma_ray_fsr`` and compute spectra::

        msqrds = Msqrds(mx=200.0, c1=1.0, lam=1e6)
        num_ps_pts = 10**6
        isp_masses = np.array([msqrds.mx, msqrds.mx])
        fsp_masses = np.array([mpi, mpi, 0.0])
        cme = 2.0 * msqrds.mx * (1 + 0.5 * 1e-6)
        num_bins = 150
        gamma_ray_fsr(isp_masses, fsp_masses, cme, msqrds.tree,
                      msqrds.radiative, num_ps_pts, num_bins)

    """
    if len(isp_masses) == 1:
        cross_section = rambo.compute_decay_width(
            num_ps_pts, fsp_masses[0:-1], cme, mat_elem_sqrd=mat_elem_sqrd_tree
        )[0]

        pre_factor = 1.0 / (2.0 * cme)
    else:
        cross_section = rambo.compute_annihilation_cross_section(
            isp_masses,
            fsp_masses[0:-1],
            cme,
            num_ps_pts=num_ps_pts,
            mat_elem_sqrd=mat_elem_sqrd_tree,
        )[0]

        m1 = isp_masses[0]
        m2 = isp_masses[1]
        pre_factor = cross_section_prefactor(m1, m2, cme)

    eng_hists = rambo.generate_energy_histogram(
        fsp_masses,
        cme,
        num_ps_pts=num_ps_pts,
        mat_elem_sqrd=mat_elem_sqrd_rad,
        num_bins=num_bins,
    )[0]

    photon_energies = eng_hists[-1, 0]
    spectrum = eng_hists[-1, 1] * pre_factor / cross_section

    return photon_energies, spectrum



"""
TODO: Update documentation.
"""
from hazma.decay_helper_functions cimport decay_charged_pion
from hazma.decay_helper_functions cimport decay_neutral_pion
import numpy as np
cimport numpy as np
from scipy.integrate import quad
from libc.math cimport sqrt
import cython
import os
import sys
from .get_path import get_dir_path
include "parameters.pxd"
import warnings


"""
Module for computing the photon spectrum from radiative short kaon decay.

Description:
    The short kaon has many decay modes:

        ks -> pi + pi
        ks -> pi0 + pi0
        ks -> pi + pi + g

    For the the two-body final states, the sum of the decay spectra are
    computed given the known energies of the final state particles in the
    kaon's rest frame. The spectrum is then boosted into the lab frame.
"""

data_path_total = os.path.join(get_dir_path(),
                               "interpolation_data",
                               "skaon",
                               "short_kaon_interp_total.dat")
data_path_00 = os.path.join(get_dir_path(),
                            "interpolation_data",
                            "skaon",
                            "short_kaon_interp_00.dat")
data_path_pm = os.path.join(get_dir_path(),
                            "interpolation_data",
                            "skaon",
                            "short_kaon_interp_pm.dat")

data_path_pmg = os.path.join(get_dir_path(),
                             "interpolation_data",
                             "skaon",
                             "short_kaon_interp_pmg.dat")

__e_gams_total, __spec_total = np.loadtxt(data_path_total, delimiter=',').T
__e_gams_00, __spec_00 = np.loadtxt(data_path_00, delimiter=',').T
__e_gams_pm, __spec_pm = np.loadtxt(data_path_pm, delimiter=',').T
__e_gams_pmg, __spec_pmg = np.loadtxt(data_path_pmg, delimiter=',').T

cdef double __interp_spec(double eng_gam, str mode):
    """
    Intepolation function for the short kaon data.

    Parameters
    ----------
    eng_gam : double
        Energy of the photon.
    mode : str {"total"}
        String specifying which decay mode to use.
    """
    if mode == "total":
        return np.interp(eng_gam, __e_gams_total, __spec_total)
    if mode == "00":
        return np.interp(eng_gam, __e_gams_00, __spec_00)
    if mode == "pm":
        return np.interp(eng_gam, __e_gams_pm, __spec_pm)
    if mode == "pmg":
        return np.interp(eng_gam, __e_gams_pmg, __spec_pmg)


@cython.cdivision(True)
cdef double __integrand(double cl, double eng_gam, double eng_k, str mode):
    """
    Integrand for K_S -> X, where X is a any final state.

    The X's used are "pi + pi", "pi0 + pi0" and "pi + pi + g".

    Parameters
    ----------
    cl: float
        Angle of photon w.r.t. charged kaon in lab frame.
    eng_gam: float
        Energy of photon in laboratory frame.
    eng_k: float
        Energy of kaon in laboratory frame.
    mode : str {"total"}
        String specifying which decay mode to use.
    """
    cdef double gamma_k = eng_k / MASS_K
    cdef double beta_k = sqrt(1.0 - (MASS_K / eng_k)**2)
    cdef double eng_gam_k_rf = eng_gam * gamma_k * (1.0 - beta_k * cl)

    cdef int i, j
    cdef double ret_val
    cdef double pre_factor \
        = 1.0 / (2.0 * gamma_k * (1.0 - beta_k * cl))

    return pre_factor * __interp_spec(eng_gam_k_rf, mode)


cdef double CSpectrumPoint(double eng_gam, double eng_k, str mode):
    """
    Returns the radiative spectrum value from charged kaon at
    a single gamma ray energy.

    Parameters
    ----------
    eng_gam : float
        Energy of photon is laboratory frame.
    eng_k : float
        Energy of charged kaon in laboratory frame.
    mode : str {"total"}
        String specifying which decay mode to use.
    """
    message = 'Energy of kaon cannot be less than the kaon mass. Returning 0.'
    if eng_k < MASS_K0:
        # raise warnings.warn(message, RuntimeWarning)
        return 0.0

    cdef double result = 0.0

    return quad(__integrand, -1.0, 1.0, points=[-1.0, 1.0], \
                  args=(eng_gam, eng_k, mode), epsabs=10**-10., \
                  epsrel=10**-4.)[0]


@cython.boundscheck(False)
@cython.wraparound(False)
cdef np.ndarray CSpectrum(np.ndarray[np.float64_t, ndim=1] eng_gams,
                          double eng_k, str mode):
    """
    Returns the radiative spectrum dNde from charged kaon for a
    list of gamma ray energies.

    Parameters
    ----------
    eng_gams : numpy.ndarray
        List of energies of photon in laboratory frame.
    eng_k : float
        Energy of charged kaon in laboratory frame.
    mode : str {"total"}
        String specifying which decay mode to use.
    """
    cdef int numpts = len(eng_gams)

    cdef np.ndarray spec = np.zeros(numpts, dtype=np.float64)

    cdef int i = 0

    for i in range(numpts):
        spec[i] = CSpectrumPoint(eng_gams[i], eng_k, mode)

    return spec


def SpectrumPoint(double eng_gam, double eng_k, str mode):
    """
    Returns the radiative spectrum value from charged kaon at
    a single gamma ray energy.

    Parameters
    ----------
    eng_gam : float
        Energy of photon is laboratory frame.
    eng_k : float
        Energy of charged kaon in laboratory frame.
    mode : str {"total"}
        String specifying which decay mode to use.
    """
    return CSpectrumPoint(eng_gam, eng_k, mode)


@cython.boundscheck(False)
@cython.wraparound(False)
def Spectrum(np.ndarray[np.float64_t, ndim=1] eng_gams, double eng_k,
             str mode):
    """
    Returns the radiative spectrum dNde from charged kaon for a
    list of gamma ray energies.

    Parameters
    ----------
    eng_gams : numpy.ndarray
        List of energies of photon in laboratory frame.
    eng_k :
        Energy of charged kaon in laboratory frame.
    mode : str {"total"}
        String specifying which decay mode to use.
    """
    return CSpectrum(eng_gams, eng_k, mode)



"""
This file contains parameters of the standard model for leptons and meson.

NOTE: All dimensionful parameters are in MeV.
"""

# MASSES (MeV)
cdef double MASS_E = 0.510998928   # electron
cdef double MASS_MU = 105.6583715  # muon
cdef double MASS_PI0 = 134.9766    # neutral pion
cdef double MASS_PI = 139.57018    # Charged pion
cdef double MASS_K0 = 497.61       # neutral kaon
cdef double MASS_K = 493.68        # charged Kaon
cdef double MASS_ETA = 547.86      # eta
cdef double MASS_ETAP = 957.8      # eta prime
cdef double MASS_RHO = 775.3       # rho
cdef double MASS_OMEGA = 782.7     # omega

# BRANCHING RATIOS
cdef double BR_PI0_TO_GG = 0.9882         # Pi0   -> g   + g
cdef double BR_PI_TO_MUNU = 0.9998        # pi    -> mu  + nu
cdef double BR_PI_TO_ENU = 1.2 * 10**-4   # pi    -> e  + nu

cdef double BR_KS_TO_PIPI = 0.6920        # ks    -> pi  + pi
cdef double BR_KS_TO_PI0PI0 = 0.3069      # ks    -> pi0 + pi0

cdef double BR_KL_TO_PIENU = 0.4055       # kl    -> pi  + e   + nu
cdef double BR_KL_TO_PIMUNU = 0.2704      # kl    -> pi  + mu  + nu
cdef double BR_KL_TO_3PI0 = 0.1952        # kl    -> pi0 + pi0  + pi0
cdef double BR_KL_TO_2PIPI0 = 0.1254      # kl    -> pi  + pi  + pi0

cdef double BR_K_TO_MUNU = 0.6356         # k     -> mu  + nu
cdef double BR_K_TO_PIPI0 = 0.2067        # k     -> pi  + pi0
cdef double BR_K_TO_3PI = 0.05583         # k     -> pi  + pi  + pi
cdef double BR_K_TO_PI0ENU = 0.0507       # k     -> pi0 + e   + nu
cdef double BR_K_TO_PI0MUNU = 0.03352     # k     -> pi0 + mu  + nu
cdef double BR_K_TO_PI2PI0 = 0.01760      # k     -> pi  + pi0 + pi0

cdef double BR_ETA_TO_GG = 0.3941         # eta   -> g   + g
cdef double BR_ETA_TO_3PI0 = 0.3268       # eta   -> pi0 + pi0 + pi0
cdef double BR_ETA_TO_2PIPI0 = 0.2292     # eta   -> pi  + pi  + pi0
cdef double BR_ETA_TO_2PIG = 0.0422       # eta   -> pi  + pi  + g
cdef double BR_ETAP_TO_2PIETA = 0.429     # eta'  -> pi  + pi  + eta
cdef double BR_ETAP_TO_RHOG = 0.291       # eta'  -> rho + g
cdef double etap_BR_pi0_pi0_eta = 0.222   # eta'  -> pi0 + pi0 + eta
cdef double BR_ETAP_TO_OMEGAG = 0.0275    # eta'  -> omega + g
cdef double BR_ETAP_TO_GG = 0.0220        # eta'  -> g   + g
cdef double BR_ETAP_TO_3PI0 = 0.0214      # eta'  -> pi0 + pi0 + pi-
cdef double BR_ETAP_TO_MUMUG = 0.0108     # eta'  -> mu  + mu  + g

cdef double BR_OMEGA_TO_2PIPI0 = 0.892    # omega -> pi + pi   + pi0
cdef double BR_OMEGA_TO_PI0G = 0.0828     # omega -> pi0 + g
cdef double BR_OMEGA_TO_2PI = 0.0153      # omega -> pi + pi

# WIDTHS
cdef double WIDTH_K = 3.3406**-13.
cdef double WIDTH_PI = 2.528511206475808**-14.

# MISC.
cdef double ALPHA_EM = 1.0 / 137.0  # Fine structure constant.
cdef double RATIO_E_MU_MASS_SQ = (MASS_E / MASS_MU)**2.
cdef double A_PI = 0.0119
cdef double V_PI = 0.0254
cdef double A_K = 0.042
cdef double V_K = 0.096
cdef double DECAY_CONST_PI = 130.41
cdef double DECAY_CONST_K = 156.1



import numpy as np

from hazma.parameters import vh, b0, alpha_em, fpi
from hazma.parameters import muon_mass as mmu
from hazma.parameters import electron_mass as me
from hazma.parameters import up_quark_mass as muq
from hazma.parameters import down_quark_mass as mdq
from hazma.parameters import charged_pion_mass as mpi

from scipy.integrate import quad


class PseudoScalarMediatorCrossSections:
    def sigma_xx_to_p_to_ff(self, Q, f):
        """Returns the cross section for two DM particles into two fermions.

        Parameters
        ----------
        Q : float
            Center of mass energy.
        f : string
            Name of final state fermions: 'e' or 'mu'.
        self : PseudoScalarMediatorParameters
            Object of the pseudoscalar parameters class.

        Returns
        -------
        cross_section : float
            Cross section for xx -> p -> ff.
        """
        if f == "e":
            rf = me / Q
            gpff = self.gpee
        elif f == "mu":
            rf = mmu / Q
            gpff = self.gpmumu

        beta = self.beta
        gpxx = self.gpxx
        mp = self.mp
        width_p = self.width_p
        rx = self.mx / Q

        if 2.0 * rf < 1 and 2.0 * rx < 1:
            ret = (
                (1 - 2 * beta ** 2)
                * gpff ** 2
                * gpxx ** 2
                * Q ** 2
                * np.sqrt(1 - 4 * rf ** 2)
            ) / (
                16.0
                * np.pi
                * np.sqrt(1 - 4 * rx ** 2)
                * ((mp ** 2 - Q ** 2) ** 2 + mp ** 2 * width_p ** 2)
            )

            assert ret.imag == 0
            assert ret.real >= 0

            return ret.real
        else:
            return 0.0

    def sigma_xx_to_p_to_gg(self, Q):
        """Returns the cross section for DM annihilating into two photons.

        Parameters
        ----------
        Q : float
            Center of mass energy.

        Returns
        -------
        cross_section : float
            Cross section for xx -> p -> gg.
        """
        beta = self.beta
        mx = self.mx

        if Q >= 2.0 * mx:
            gpFF = self.gpFF
            gpxx = self.gpxx
            mp = self.mp
            rx = mx / Q
            width_p = self.width_p

            ret = (
                alpha_em ** 2
                * gpxx ** 2
                * Q ** 4
                * (
                    (1 - 2 * beta ** 2) * fpi ** 2 * gpFF ** 2
                    - 2 * beta * fpi * gpFF * vh
                    + beta ** 2 * vh ** 2
                )
            ) / (
                256.0
                * fpi ** 2
                * np.pi ** 3
                * np.sqrt(1.0 - 4 * rx ** 2)
                * vh ** 2
                * ((mp ** 2 - Q ** 2) ** 2 + mp ** 2 * width_p ** 2)
            )

            assert ret.imag == 0
            assert ret.real >= 0

            return ret
        else:
            return 0.0

    def sigma_xx_to_pp(self, Q):
        """Returns the cross section for DM annihilating into two mediators.

        Parameters
        ----------
        Q : float
            Center of mass energy.
        self : PseudoScalarMediatorParameters
            Object of the pseudoscalar parameters class.

        Returns
        -------
        cross_section : float
            Cross section for xx -> pp.
        """
        mx = self.mx
        mp = self.mp
        beta = self.beta

        if Q > 2.0 * mp and Q >= 2.0 * mx:
            gpxx = self.gpxx
            rp = mp / Q
            rx = mx / Q

            ret = (
                (-1 + 2 * beta ** 2)
                * gpxx ** 4
                * (
                    (
                        2
                        * np.sqrt((-1 + 4 * rp ** 2) * (-1 + 4 * rx ** 2))
                        * (3 * rp ** 4 + 2 * rx ** 2 - 8 * rp ** 2 * rx ** 2)
                    )
                    / (rp ** 4 + rx ** 2 - 4 * rp ** 2 * rx ** 2)
                    + (
                        2
                        * (1 - 4 * rp ** 2 + 6 * rp ** 4)
                        * (
                            1j * np.pi
                            + 2
                            * np.arctanh(
                                (1 - 2 * rp ** 2)
                                / np.sqrt((-1 + 4 * rp ** 2) * (-1 + 4 * rx ** 2))
                            )
                        )
                    )
                    / (-1 + 2 * rp ** 2)
                )
            ) / (64.0 * Q ** 2 * np.pi * (1 - 4 * rx ** 2))

            assert ret.imag == 0
            assert ret.real >= 0

            return ret
        else:
            return 0.0

    def dsigma_ds_xx_to_p_to_pi0pi0pi0(self, s, Q):
        """Returns the dsigma/ds for DM annihilation into three neutral pions.

        Parameters
        ----------
        Q : float
            Center of mass energy.

        Returns
        -------
        dsigma_ds : float
            dsigma/ds for xx -> p -> pi0 pi0 pi0, where s=(P-q)^2, with P the
            center of mass momentum and q the momentum of one of the pi0s.
        """
        mx = self.mx
        mpi0 = self.mpi0  # use shifted pion mass!

        if Q > 2.0 * mx and Q >= 3.0 * mpi0:
            beta = self.beta
            gpxx = self.gpxx
            gpuu = self.gpuu
            gpdd = self.gpdd
            gpGG = self.gpGG
            mp = self.mp
            width_p = self.width_p

            ret = -(
                b0 ** 2
                * gpxx ** 2
                * np.sqrt(s * (-4 * mpi0 ** 2 + s))
                * np.sqrt(mpi0 ** 4 + (Q ** 2 - s) ** 2 - 2 * mpi0 ** 2 * (Q ** 2 + s))
                * (
                    -(beta ** 2 * (mdq + muq) ** 2 * vh ** 2)
                    + 2
                    * beta
                    * fpi
                    * (mdq + muq)
                    * vh
                    * (gpGG * (mdq - muq) + (gpdd - gpuu) * vh)
                    + (-1 + 11 * beta ** 2)
                    * fpi ** 2
                    * (gpGG * (mdq - muq) + (gpdd - gpuu) * vh) ** 2
                )
            ) / (
                512.0
                * fpi ** 4
                * np.pi ** 3
                * Q
                * np.sqrt(-4 * mx ** 2 + Q ** 2)
                * s
                * vh ** 2
                * (mp ** 4 + Q ** 4 + mp ** 2 * (-2 * Q ** 2 + width_p ** 2))
            )

            assert ret.imag == 0
            assert ret.real >= 0

            return ret
        else:
            return 0.0

    def sigma_xx_to_p_to_pi0pi0pi0(self, Q):
        """Returns the DM annihilation cross section into three neutral pions.

        Notes
        -----
        Integrates dsigma/ds as given by `dsigma_ds_xx_to_p_to_pi0pi0pi0`
        over s.

        Parameters
        ----------
        Q : float
            Center of mass energy.
        self : PseudoScalarMediatorParameters
            Object of the pseudoscalar parameters class.

        Returns
        -------
        cross_section : float
            The DM annihilation cross section xx -> p -> pi0 pi0 pi0.
        """

        mpi0 = self.mpi0  # use shifted pion mass!
        smax = (Q - mpi0) ** 2
        smin = 4.0 * mpi0 ** 2

        res = quad(self.dsigma_ds_xx_to_p_to_pi0pi0pi0, smin, smax, args=(Q))

        return res[0]

    def dsigma_ds_xx_to_p_to_pi0pipi(self, s, Q):
        """Returns the dsigma/ds for DM annihilation into a neutral pion and two
        charged pions.

        Parameters
        ----------
        Q : float
            Center of mass energy.
        self : PseudoScalarMediatorParameters
            Object of the pseudoscalar parameters class.

        Returns
        -------
        dsigma_ds : float
            dsigma/ds for xx -> p -> pi0 pi pi, where s=(P-q)^2, with P the
            center of mass momentum and q the momentum of the pi0.
        """
        mx = self.mx
        mpi0 = self.mpi0  # use shifted pion mass!

        if Q > 2.0 * mx and Q >= 2.0 * mpi + mpi0:
            beta = self.beta
            gpxx = self.gpxx
            gpuu = self.gpuu
            gpdd = self.gpdd
            gpGG = self.gpGG
            mp = self.mp
            width_p = self.width_p

            ret = (
                gpxx ** 2
                * np.sqrt(s * (-4 * mpi ** 2 + s))
                * np.sqrt(mpi0 ** 4 + (Q ** 2 - s) ** 2 - 2 * mpi0 ** 2 * (Q ** 2 + s))
                * (
                    beta ** 2 * (2 * mpi ** 2 + mpi0 ** 2 - 3 * s) ** 2 * vh ** 2
                    + 2
                    * b0
                    * beta
                    * (2 * mpi ** 2 + mpi0 ** 2 - 3 * s)
                    * vh
                    * (
                        -(beta * (mdq + muq) * vh)
                        + fpi * (gpGG * (mdq - muq) + (gpdd - gpuu) * vh)
                    )
                    + b0 ** 2
                    * (
                        beta ** 2 * (mdq + muq) ** 2 * vh ** 2
                        - 2
                        * beta
                        * fpi
                        * (mdq + muq)
                        * vh
                        * (gpGG * (mdq - muq) + (gpdd - gpuu) * vh)
                        - (-1 + 5 * beta ** 2)
                        * fpi ** 2
                        * (gpGG * (mdq - muq) + (gpdd - gpuu) * vh) ** 2
                    )
                )
            ) / (
                4608.0
                * fpi ** 4
                * np.pi ** 3
                * Q
                * np.sqrt(-4 * mx ** 2 + Q ** 2)
                * s
                * vh ** 2
                * (mp ** 4 + Q ** 4 + mp ** 2 * (-2 * Q ** 2 + width_p ** 2))
            )

            assert ret.imag == 0
            assert ret.real >= 0

            return ret
        else:
            return 0.0

    def sigma_xx_to_p_to_pi0pipi(self, Q):
        """Returns the DM annihilation cross section into a neutral pion and two
        charged pions.

        Notes
        -----
        Integrates dsigma/ds as given by `dsigma_ds_xx_to_p_to_pi0pipi` over s.

        Parameters
        ----------
        Q : float
            Center of mass energy.
        self : PseudoScalarMediatorParameters
            Object of the pseudoscalar parameters class.

        Returns
        -------
        cross_section : float
            The DM annihilation cross section xx -> p -> pi0 pi pi.
        """

        mpi0 = self.mpi0  # use shifted pion mass!
        smax = (Q - mpi0) ** 2
        smin = 4.0 * mpi ** 2

        res = quad(self.dsigma_ds_xx_to_p_to_pi0pipi, smin, smax, args=(Q))

        return res[0]

    def annihilation_cross_sections(self, Q):
        """
        Compute the total cross section DM annihilation.

        Parameters
        ----------
        Q : float
            Center of mass energy.

        Returns
        -------
        cs : dict
            Dictionary containing the theory's cross sections. The keys are
            'total', 'mu mu', 'e e', 'pi0 pi0 pi0', 'pi0 pi pi', 'g g', 'p p'.
        """
        muon_contr = self.sigma_xx_to_p_to_ff(Q, "mu")
        electron_contr = self.sigma_xx_to_p_to_ff(Q, "e")
        photon_contr = self.sigma_xx_to_p_to_gg(Q)

        pi0pipi_contr = self.sigma_xx_to_p_to_pi0pipi(Q)
        pi0pi0pi0_contr = self.sigma_xx_to_p_to_pi0pi0pi0(Q)

        pp_contr = self.sigma_xx_to_pp(Q)

        total = (
            muon_contr
            + electron_contr
            + pi0pipi_contr
            + pi0pi0pi0_contr
            + photon_contr
            + pp_contr
        )

        cross_secs = {
            "mu mu": muon_contr,
            "e e": electron_contr,
            "pi0 pi pi": pi0pipi_contr,
            "pi0 pi0 pi0": pi0pi0pi0_contr,
            "g g": photon_contr,
            "p p": pp_contr,
            "total": total,
        }

        return cross_secs

    def annihilation_branching_fractions(self, Q):
        """
        Compute the branching fractions for DM annihilation.

        Parameters
        ----------
        Q : float
            Center of mass energy.

        Returns
        -------
        bfs : dictionary
            Dictionary of the branching fractions. The keys are 'total',
            'mu mu', 'e e', 'pi0 pi0 pi0', 'pi0 pi pi', 'g g', 'p p'.
        """
        CSs = self.annihilation_cross_sections(Q)

        bfs = {
            "e e": CSs["e e"] / CSs["total"],
            "mu mu": CSs["mu mu"] / CSs["total"],
            "g g": CSs["g g"] / CSs["total"],
            "p p": CSs["p p"] / CSs["total"],
            "pi0 pi pi": CSs["pi0 pi pi"] / CSs["total"],
            "pi0 pi0 pi0": CSs["pi0 pi0 pi0"] / CSs["total"],
        }

        return bfs



import numpy as np
cimport numpy as np

cdef np.ndarray __e_gams_total
cdef np.ndarray __e_gams_0enu
cdef np.ndarray __e_gams_0munu
cdef np.ndarray __e_gams_00p
cdef np.ndarray __e_gams_mmug
cdef np.ndarray __e_gams_munu
cdef np.ndarray __e_gams_p0
cdef np.ndarray __e_gams_p0g
cdef np.ndarray __e_gams_ppm

cdef np.ndarray __spec_total
cdef np.ndarray __spec_0enu
cdef np.ndarray __spec_0munu
cdef np.ndarray __spec_00p
cdef np.ndarray __spec_mmug
cdef np.ndarray __spec_munu
cdef np.ndarray __spec_p0
cdef np.ndarray __spec_p0g
cdef np.ndarray __spec_ppm

cdef double __interp_spec(double, str)

cdef double __integrand(double, double, double, str)

cdef double CSpectrumPoint(double, double, str)

cdef np.ndarray CSpectrum(np.ndarray[np.float64_t, ndim=1], double, str)



import numpy as np
from cmath import sqrt, pi, atanh, log

from hazma.parameters import qe, alpha_em, charged_pion_mass as mpi


class ScalarMediatorFSR:
    def __dnde_xx_to_s_to_ffg(self, photon_energy, Q, mf):
        """ Unvectorized dnde_xx_to_s_to_ffg """
        e, rf, s = photon_energy / Q, mf / Q, Q ** 2 - 2.0 * Q * photon_energy

        mx = self.mx

        if 2.0 * mf < Q and 4.0 * mf ** 2 < s < Q ** 2 and 2.0 * mx < Q:
            ret_val = (
                alpha_em
                * (
                    2
                    * (-1 + 4 * rf ** 2)
                    * sqrt((-1 + 2 * e) * (-1 + 2 * e + 4 * rf ** 2))
                    + 4
                    * (
                        1
                        + 2 * (-1 + e) * e
                        - 6 * rf ** 2
                        + 8 * e * rf ** 2
                        + 8 * rf ** 4
                    )
                    * atanh(sqrt(1 + (4 * rf ** 2) / (-1 + 2 * e)))
                )
            ) / (e * (1 - 4 * rf ** 2) ** 1.5 * pi * Q)

            assert ret_val.imag == 0
            ret_val = ret_val.real
            assert ret_val >= 0

            return ret_val
        else:
            return 0.0

    def dnde_xx_to_s_to_ffg(self, photon_energies, Q, mf):
        """Return the fsr spectra for dark matter annihilating into a pair of
        fermions.

        Computes the final state radiation spectrum value dNdE from a scalar
        mediator given a gamma ray energy of `eng_gam`, center of mass
        energy `cme`
        and final state fermion mass `mass_f`.

        Parameters
        ----------
        photon_energies : float or np.array
            Energy(ies) of the final state photon.
        Q: float
            Center of mass energy of mass of off-shell scalar mediator.
        mf : float
            Mass of the final state fermion.

        Returns
        -------
        spec_val : float
            Spectrum value dNdE from scalar mediator.

        """
        if hasattr(photon_energies, "__len__"):
            return np.array(
                [self.__dnde_xx_to_s_to_ffg(e, Q, mf) for e in photon_energies]
            )
        else:
            return self.__dnde_xx_to_s_to_ffg(photon_energies, Q, mf)

    def __dnde_xx_to_s_to_pipig(self, photon_energy, Q):
        """Unvectorized dnde_xx_to_s_to_pipig"""

        mupi = mpi / Q
        x = 2 * photon_energy / Q

        if x < 0.0 or 1.0 - 4.0 * mupi ** 2 < x or 2.0 * self.mx > Q:
            return 0.0

        dynamic = (
            2 * sqrt(-1 + x) * sqrt(-1 + 4 * mupi ** 2 + x)
            + (-1 + 2 * mupi ** 2 + x)
            * log(
                (1 - x + sqrt(-1 + x) * sqrt(-1 + 4 * mupi ** 2 + x)) ** 2
                / (-1 + x + sqrt(-1 + x) * sqrt(-1 + 4 * mupi ** 2 + x)) ** 2
            )
        ) / x

        coeff = qe ** 2 / (4.0 * sqrt(1 - 4 * mupi ** 2) * pi ** 2)

        ret_val = 2.0 * dynamic * coeff / Q

        assert ret_val.imag == 0.0
        assert ret_val.real >= 0.0

        return ret_val.real

    def dnde_xx_to_s_to_pipig(self, photon_energies, Q):
        """
        Returns the gamma ray energy spectrum for two fermions annihilating
        into two charged pions and a photon.

        Parameters
        ----------
        photon_energies : float or np.array
            Energy(ies) of the final state photon.
        Q : double
            Center of mass energy

        Returns
        -------
        dnde: float or np.array
            Gamma-ray spectrum for dark matter annihilating into charged pions.

        """

        if hasattr(photon_energies, "__len__"):
            return np.array(
                [
                    self.__dnde_xx_to_s_to_pipig(eng_gam, Q)
                    for eng_gam in photon_energies
                ]
            )
        else:
            return self.__dnde_xx_to_s_to_pipig(photon_energies, Q)



from hazma import gamma_ray
import warnings
import numpy as np

import unittest

warnings.filterwarnings("ignore")


class TestGammaRay(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        pass

    @classmethod
    def tearDownClass(cls):
        pass

    def setUp(self):
        self.cme = 5000.0
        self.eng_gams = np.logspace(0.0, np.log10(5000), num=10, dtype=np.float64)

    def tearDown(self):
        pass

    def test_mu_mu_mu(self):
        particles = np.array(["muon", "muon", "muon"])
        gamma_ray.gamma_ray_decay(particles, self.cme, self.eng_gams)

    def test_ck_mu_mu(self):
        particles = np.array(["charged_kaon", "muon", "muon"])
        gamma_ray.gamma_ray_decay(particles, self.cme, self.eng_gams)

    def test_sk_mu_mu(self):
        particles = np.array(["short_kaon", "muon", "muon"])
        gamma_ray.gamma_ray_decay(particles, self.cme, self.eng_gams)

    def test_lk_mu_mu(self):
        particles = np.array(["long_kaon", "muon", "muon"])
        gamma_ray.gamma_ray_decay(particles, self.cme, self.eng_gams)

    def test_cp_mu_mu(self):
        particles = np.array(["charged_pion", "muon", "muon"])
        gamma_ray.gamma_ray_decay(particles, self.cme, self.eng_gams)

    def test_np_mu_mu(self):
        particles = np.array(["neutral_pion", "muon", "muon"])
        gamma_ray.gamma_ray_decay(particles, self.cme, self.eng_gams)



import numpy as np
cimport numpy as np

cdef np.ndarray eng_ps_mu

cdef double __muon_spectrum(double)
cdef double __integrand(double, double, double)
cdef double CSpectrumPoint(double, double)
cdef np.ndarray CSpectrum(np.ndarray, double)



"""
Module for computing positron spectra.

@author: Logan Morrison and Adam Coogan
@date: May 2018
"""

import numpy as np

from hazma.positron_helper_functions import positron_charged_pion, positron_muon
from hazma.positron_helper_functions.positron_decay import positron, positron_point


def muon(positron_energies, muon_energy):
    """
    Returns the positron spectrum from muon decay.

    Parameters
    ----------
    positron_energies : float or numpy.array
        Energy(ies) of the positron/electron.
    muon_energy : float or array-like
        Energy of the muon.

    Returns
    -------
    dnde : float or numpy.array
        The value of the spectrum given a positron energy(ies)
        ``positron_energies`` and muon energy ``muon_energy``.
    """
    if hasattr(positron_energies, "__len__"):
        return positron_muon.Spectrum(positron_energies, muon_energy)
    return positron_muon.SpectrumPoint(positron_energies, muon_energy)


def charged_pion(positron_energies, pion_energy):
    """
    Returns the positron spectrum from the decay of a charged pion.

    Parameters
    ----------
    positron_energies : float or numpy.array
        Energy(ies) of the positron/electron.
    pion_energy : float or numpy.array
        Energy of the charged pion.

    Returns
    -------
    dnde : float or numpy.array
        The value of the spectrum given a positron energy(ies)
        ``positron_energies`` and charged pion energy ``pion_energy``.
    """
    if hasattr(positron_energies, "__len__"):
        return positron_charged_pion.Spectrum(positron_energies, pion_energy)
    return positron_charged_pion.SpectrumPoint(positron_energies, pion_energy)


def positron_decay(
    particles,
    cme,
    positron_energies,
    mat_elem_sqrd=lambda k_list: 1.0,
    num_ps_pts=1000,
    num_bins=25,
    verbose=False,
):
    r"""Returns total gamma ray spectrum from a set of particles.

    Parameters
    ----------
    particles : array_like
        List of particle names. Available particles are 'muon', and
        'charged_pion'.
    cme : double
        Center of mass energy of the final state in MeV.
    positron_energies : np.ndarray[double, ndim=1]
        List of positron energies in MeV to evaluate spectra at.
    mat_elem_sqrd : double(\*func)(np.ndarray)
        Function for the matrix element squared of the process. Must be a
        function taking in a list of four momenta of size (num_fsp, 4).
        Default value is a flat matrix element returning 1..
    num_ps_pts : int {1000}, optional
        Number of phase space points to use.
    num_bins : int {25}, optional
        Number of bins to use.
    verbose: Bool
        If True, then addition information about the runtime progress and state
        are displayed.

    Returns
    -------
    spec : np.ndarray
        Total gamma ray spectrum from all final state particles.

    Notes
    -----
    The total spectrum is computed using

    .. math::
        \frac{dN}{dE}(E_{e^{\pm}}) = \sum_{i,j}P_{i}(E_{j})
        \frac{dN_i}{dE}(E_{e^{\pm}}, E_{j})

    where :math:`i` runs over the final state particles, :math:`j` runs over
    energies sampled from probability distributions. :math:`P_{i}(E_{j})` is
    the probability that particle :math:`i` has energy :math:`E_{j}`. The
    probabilities are computed using ``hazma.phase_space_generator.rambo``. The
    total number of energies used is ``num_bins``.

    Examples
    --------

    Generate spectrum from a muon, and two charged pions
    with total energy of 5 GeV::

        from hazma.positron_spectra import positron_decay
        from hazma.parameters import electron_mass as me
        import numpy as np
        particles = np.array(['muon', 'charged_pion', 'charged_pion'])
        cme = 5000.
        positron_energies = np.logspace(np.log10(me), np.log10(cme),
                                        num=200, dtype=np.float64)
        positron_decay(particles, cme, positron_energies)

    """
    if type(particles) == str:
        particles = [particles]

    particles = np.array(particles)

    if hasattr(positron_energies, "__len__"):
        return positron(
            particles,
            cme,
            positron_energies,
            mat_elem_sqrd=mat_elem_sqrd,
            num_ps_pts=num_ps_pts,
            num_bins=num_bins,
            verbose=verbose,
        )
    return positron_point(
        particles, cme, positron_energies, mat_elem_sqrd, num_ps_pts, num_bins
    )



from hazma.theory import Theory

from hazma.parameters import electron_mass as me
from hazma.parameters import muon_mass as mmu
from hazma.parameters import up_quark_mass as muq
from hazma.parameters import down_quark_mass as mdq
from hazma.parameters import strange_quark_mass as msq
from hazma.parameters import vh

from hazma.parameters import b0, fpi
from hazma.parameters import neutral_pion_mass as mpi0

from hazma.pseudo_scalar_mediator._pseudo_scalar_mediator_cross_sections import (
    PseudoScalarMediatorCrossSections,
)
from hazma.pseudo_scalar_mediator._pseudo_scalar_mediator_fsr import (
    PseudoScalarMediatorFSR,
)
from hazma.pseudo_scalar_mediator._pseudo_scalar_mediator_msqrd_rambo import (
    PseudoScalarMediatorMSqrdRambo,
)
from hazma.pseudo_scalar_mediator._pseudo_scalar_mediator_positron_spectra import (
    PseudoScalarMediatorPositronSpectra,
)
from hazma.pseudo_scalar_mediator._pseudo_scalar_mediator_spectra import (
    PseudoScalarMediatorSpectra,
)
from hazma.pseudo_scalar_mediator._pseudo_scalar_mediator_widths import (
    PseudoScalarMediatorWidths,
)

import warnings
from hazma.hazma_errors import PreAlphaWarning

import numpy as np


# Note that Theory must be inherited from AFTER all the other mixin classes,
# since they furnish definitions of the abstract methods in Theory.
class PseudoScalarMediator(
    PseudoScalarMediatorCrossSections,
    PseudoScalarMediatorFSR,
    PseudoScalarMediatorMSqrdRambo,
    PseudoScalarMediatorPositronSpectra,
    PseudoScalarMediatorSpectra,
    PseudoScalarMediatorWidths,
    Theory,
):
    r"""
    Create a pseudoscalar mediator model object.
    """

    def __init__(self, mx, mp, gpxx, gpuu, gpdd, gpss, gpee, gpmumu, gpGG, gpFF):
        self._mx = mx
        self._mp = mp
        self._gpxx = gpxx
        self._gpuu = gpuu
        self._gpdd = gpdd
        self._gpss = gpss
        self._gpee = gpee
        self._gpmumu = gpmumu
        self._gpGG = gpGG
        self._gpFF = gpFF

        self.determine_mixing()  # must be called BEFORE computing P's width!
        self.compute_width_p()

    @property
    def mx(self):
        return self._mx

    @mx.setter
    def mx(self, mx):
        self._mx = mx
        self.compute_width_p()

    @property
    def mp(self):
        return self._mp

    @mp.setter
    def mp(self, mp):
        self._mp = mp
        self.determine_mixing()
        self.compute_width_p()

    @property
    def gpxx(self):
        return self._gpxx

    @gpxx.setter
    def gpxx(self, gpxx):
        self._gpxx = gpxx
        self.compute_width_p()

    @property
    def gpee(self):
        return self._gpee

    @gpee.setter
    def gpee(self, gpee):
        self._gpee = gpee
        self.compute_width_p()

    @property
    def gpmumu(self):
        return self._gpmumu

    @gpmumu.setter
    def gpmumu(self, gpmumu):
        self._gpmumu = gpmumu
        self.compute_width_p()

    @property
    def gpuu(self):
        return self._gpuu

    @gpuu.setter
    def gpuu(self, gpuu):
        self._gpuu = gpuu
        self.determine_mixing()
        self.compute_width_p()

    @property
    def gpdd(self):
        return self._gpdd

    @gpdd.setter
    def gpdd(self, gpdd):
        self._gpdd = gpdd
        self.determine_mixing()
        self.compute_width_p()

    @property
    def gpss(self):
        return self._gpss

    @gpss.setter
    def gpss(self, gpss):
        self._gpss = gpss
        # This impacts P-eta mixing, but not P-pi0 mixing. The eta is above the
        # mass range we are interested in.
        self.compute_width_p()

    @property
    def gpGG(self):
        return self._gpGG

    @gpGG.setter
    def gpGG(self, gpGG):
        self._gpGG = gpGG
        self.determine_mixing()
        self.compute_width_p()

    @property
    def gpFF(self):
        return self._gpFF

    @gpFF.setter
    def gpFF(self, gpFF):
        self._gpFF = gpFF
        self.compute_width_p()

    def determine_mixing(self):
        """Recomputes the mixing between P and pi0 and computes the resulting
        shift in mpi0.

        Notes
        -----
        * Depends on gpuu, gpdd, gpGG and mp.
        * Must be called before calling compute_width_p().
        * Resets the attributes `beta` (the mixing angle) and `mpi0` (the
          shifted neutral pion mass). Warnings are printed if the neutral pion
          mass shift is larger than 10 MeV, with a specific warning if the mass
          is shifted to be imaginary.
        """
        eps = b0 * fpi * (self.gpuu - self.gpdd + (muq - mdq) / vh * self.gpGG)

        # Mixing angle between pi0 and p. Here I have assumed that the pi0 mass
        # is given by leading order chiPT.
        self.beta = eps / (self.mp ** 2 - mpi0 ** 2)

        # Shifted mass of neutral pion
        mpi0Sqrd = mpi0 ** 2 - eps * self.beta

        if mpi0Sqrd < 0:  # mixing is way too big if this fails
            print(
                "Warning: your choice of mp and/or couplings produced an"
                + " imaginary neutral pion mass. Undefined behavior."
            )

        self.mpi0 = np.sqrt(mpi0Sqrd)

        if abs(self.mpi0 - mpi0) > 10.0:
            print(
                "Warning: your choice of mp and/or couplings produced a "
                + "10 MeV or larger shift in m_pi0. Theory is invalid."
            )

    def compute_width_p(self):
        """Updates the pseudoscalar's total width.
        """
        self.width_p = self.partial_widths()["total"]

    def description(self):
        warnings.warn("", PreAlphaWarning)
        pass

    @classmethod
    def list_annihilation_final_states(cls):
        """
        Return a list of the available final states.

        Returns
        -------
        fs : array-like
            Array of the available final states.
        """
        return ["e e", "mu mu", "g g", "pi0 pi pi", "pi0 pi0 pi0", "p p"]

    def constraints(self):
        pass


class PseudoScalarMFV(PseudoScalarMediator):
    """MFV version of the pseudoscalar model.
    """

    def __init__(self, mx, mp, gpxx, gpup, gpdown, gpll):
        self._gpup = gpup
        self._gpdown = gpdown
        self._gpll = gpll

        yu = muq / vh
        yd = mdq / vh
        ys = msq / vh

        ye = me / vh
        ymu = mmu / vh

        gpGG = 2.0 * gpup + gpdown
        gpFF = gpll + (8.0 * gpup + gpdown) / 9.0

        super(PseudoScalarMFV, self).__init__(
            mx,
            mp,
            gpxx,
            gpup * yu,
            gpdown * yd,
            gpdown * ys,
            gpll * ye,
            gpll * ymu,
            gpGG,
            gpFF,
        )

    @property
    def gpll(self):
        return self._gpll

    @gpll.setter
    def gpll(self, gpll):
        self._gpll = gpll

        ye = me / vh
        ymu = mmu / vh

        self.gpee = gpll * ye
        self.gpmumu = gpll * ymu

        self.gpFF = gpll + (8.0 * self.gpup + self.gpdown) / 9.0

    @property
    def gpup(self):
        return self._gpup

    @gpup.setter
    def gpup(self, gpup):
        self._gpup = gpup

        yu = muq / vh

        self.gpuu = gpup * yu

        self.gpFF = self.gpll + (8.0 * gpup + self.gpdown) / 9.0
        self.gpGG = 2.0 * gpup + self.gpdown

    @property
    def gpdown(self):
        return self._gpdown

    @gpdown.setter
    def gpdown(self, gpdown):
        self._gpdown = gpdown

        yd = mdq / vh
        ys = msq / vh

        self.gpdd = gpdown * yd
        self.gpss = gpdown * ys

        self.gpFF = self.gpll + (8.0 * self.gpup + gpdown) / 9.0
        self.gpGG = 2.0 * self.gpup + gpdown

    def f_eff_g(self, x_kd=1e-4):
        # Need to override the default method since the pseudoscalar uses RAMBO
        # for its spectra.
        return self._f_eff_helper("g g", x_kd, "interp")

    def f_eff_ep(self, x_kd=1e-4):
        # Need to override the default method since the pseudoscalar uses RAMBO
        # for its spectra.
        return self._f_eff_helper("e e", x_kd, "interp")



"""
Module for generating a relativistic phase space point.

* Author - Logan A. Morrison and Adam Coogan.
* Date - December 2017
"""

import numpy as np
cimport numpy as np
from libc.math cimport log, M_PI, sqrt, tgamma, fabs, pow, cos, sin
from libcpp cimport bool
import cython



cdef extern from "<random>" namespace "std":
    cdef cppclass mt19937 nogil:
        mt19937() nogil
        mt19937(unsigned int seed) nogil

    cdef cppclass random_device nogil:
        random_device() except +
        unsigned int operator()()

    cdef cppclass uniform_real_distribution[T] nogil:
        uniform_real_distribution() nogil
        uniform_real_distribution(T a, T b) nogil
        T operator()(mt19937 gen) nogil

cdef uniform_real_distribution[double] uniform \
    = uniform_real_distribution[double](0., 1.)

cdef mt19937 rng

cdef extern from "<vector>" namespace "std":
    cdef cppclass vector[T]:
        void push_back(T&) nogil except+
        size_t size() nogil
        T& operator[](size_t)

cdef extern from "random" namespace "std":
    cdef cppclass random


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef double __f_xi(double xi, vector[double] masses, double cme, int num_fsp,
vector[double] ps) nogil:
    """
    Function whose zero is the scaling factor to correct masses of massless
    four momenta.

    Parameters
    ----------
    xi : double
        Scaling factor.
    masses : double[:]
        List of masses of the final state particles.
    cme : double
        Center of mass energy of the process.
    ps : double[:]
        List of the four momentum of the final state particles.

    Returns
    -------
    f(xi) : double
        Value of function evaluated at xi.
    """
    cdef int i
    cdef double val = 0.0

    for i in range(num_fsp):
        val = val + sqrt(pow(masses[i], 2.) + pow(xi * ps[4 * i], 2.))

    return val - cme

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef double __df_xi(double xi, vector[double] masses, double cme, int num_fsp,
vector[double] ps) nogil:
    """
    Derivative of the func_xi.

    Parameters
    ----------
    xi : double
        Scaling factor.
    masses : double[:]
        List of masses of the final state particles.
    cme : double
        Center of mass energy of the process.
    ps : double[:]
        List of the four momentum of the final state particles.

    Returns
    -------
    df/dxi(xi) : double
        Value of derivative of f(xi) evaluated at xi.
    """
    cdef int i
    cdef double val = 0.0
    cdef double denom

    for i in range(num_fsp):
        denom = sqrt(pow(masses[i], 2.) + pow(xi * ps[4 * i], 2.))
        val = val + xi * pow(ps[4 * i], 2.) / denom

    return val


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef double __find_root(vector[double] masses, double cme, int num_fsp, vector[double] ps, double tol=10**-4.0, int max_iter=50) nogil:
    """
    Function for finding the scaling parameter to turn massless four-vectors
    the correct set of masses.

    Parameters
    ----------
    xi : double
        Scaling factor.
    masses : double[:]
        List of masses of the final state particles.
    cme : double
        Center of mass energy of the process.
    ps : double[:]
        List of the four momentum of the final state particles.

    Returns
    -------
    xi : double
        The scaling factor.
    """
    cdef double mass_sum = 0.0
    cdef double xi0, xi1, xi2
    cdef int i, iter_count
    cdef bool isDone
    cdef double f, df

    for i in range(num_fsp):
        mass_sum = mass_sum + masses[i]

    xi0 = sqrt(1.0 - pow(mass_sum / cme, 2.0))

    isDone = False
    iter_count = 0
    xi2 = xi0

    while isDone is False:
        if iter_count > max_iter:
            break

        xi1 = xi2

        f = __f_xi(xi1, masses, cme, num_fsp, ps)
        df = __df_xi(xi1, masses, cme, num_fsp, ps)

        xi2 = xi1 - (f / df)

        if fabs(xi2-xi1) < tol:
            isDone=True
        iter_count = iter_count + 1

    return xi2


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef double __get_mass(vector[double] fv) nogil:
    """
    Computes mass of a four-vector.

    Parameters
    ----------
    fv : double[:]
        Four-vector to compute mass of.

    Returns
    -------
    mass : double
        Mass of four-vector.
    """
    return sqrt(pow(fv[0], 2.) - \
        pow(fv[1], 2.) - pow(fv[2], 2.) - pow(fv[3], 2.))


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef vector[double] __generate_qs(vector[double] masses, double cme, int num_fsp):
    """
    Computes isotropic, random four-vectors with energies, q_0, distributed
    according to q_0 * exp(-q_0).

    Parameters
    ----------
    masses : double[:]
        List of masses of the final state particles.
    cme : double
        Center of mass energy of the process.
    rands : double[:]
        List of the random numbers uniformly distributed on (0,1).

    Returns
    -------
    qs : double[:]
        List of the massless four-momenta.
    """
    cdef int i
    cdef double rho_1, rho_2, rho_3, rho_4
    cdef double c, phi
    cdef double q_e, q_x, q_y, q_z

    # cdef double[:] qs = np.empty(num_fsp * 4 + 1, dtype=np.float64)
    cdef vector[double] qs

    for _ in range(num_fsp * 4 + 1):
        qs.push_back(0)

    for i in range(num_fsp):
        rho_1 = uniform(rng)
        rho_2 = uniform(rng)
        rho_3 = uniform(rng)
        rho_4 = uniform(rng)

        c = 2.0 * rho_1 - 1.0
        phi = 2.0 * M_PI * rho_2

        q_e = -log(rho_3 * rho_4)
        q_x = q_e * sqrt(1.0 - pow(c, 2.)) * cos(phi)
        q_y = q_e * sqrt(1.0 - pow(c, 2.)) * sin(phi)
        q_z = q_e * c

        qs[4 * i + 0] = q_e
        qs[4 * i + 1] = q_x
        qs[4 * i + 2] = q_y
        qs[4 * i + 3] = q_z

    return qs


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef vector[double] __generate_ps(vector[double] masses, double cme, int num_fsp, vector[double] qs):
    """
    Generates a list of four-momentum with correct center of mass energy from
    isotropic, random four-momentum with energies, q_0,  distributed according
    to q_0 * exp(-q_0).

    Parameters
    ----------
    masses : double[:]
        List of masses of the final state particles.
    cme : double
        Center of mass energy of the process.
    qs : double[:]
        List of the random four-momenta.

    Returns
    -------
    ps : double[:]
        List of the massless four-momenta with correct center of mass energy.
    """
    cdef int i
    cdef double mass_Q
    cdef double b_x, b_y, b_z
    cdef double x, gamma, a
    cdef double qi_e, qi_x, qi_y, qi_z
    cdef double b_dot_qi
    cdef double pi_e, pi_x, pi_y, pi_z

    cdef vector[double] sum_qs

    for i in range(4):
        sum_qs.push_back(0)

    for i in range(num_fsp):
        sum_qs[0] = sum_qs[0] + qs[4 * i + 0]
        sum_qs[1] = sum_qs[1] + qs[4 * i + 1]
        sum_qs[2] = sum_qs[2] + qs[4 * i + 2]
        sum_qs[3] = sum_qs[3] + qs[4 * i + 3]

    mass_Q = __get_mass(sum_qs)

    b_x = -sum_qs[1] / mass_Q
    b_y = -sum_qs[2] / mass_Q
    b_z = -sum_qs[3] / mass_Q
    x = cme / mass_Q
    gamma = sum_qs[0] / mass_Q
    a = 1.0 / (1.0 + gamma)

    for i in range(num_fsp):
        qi_e = qs[4 * i + 0]
        qi_x = qs[4 * i + 1]
        qi_y = qs[4 * i + 2]
        qi_z = qs[4 * i + 3]

        b_dot_qi = b_x * qi_x + b_y * qi_y + b_z * qi_z

        pi_e = x * (gamma * qi_e + b_dot_qi)
        pi_x = x * (qi_x + b_x * qi_e + a * b_dot_qi * b_x)
        pi_y = x * (qi_y + b_y * qi_e + a * b_dot_qi * b_y)
        pi_z = x * (qi_z + b_z * qi_e + a * b_dot_qi * b_z)

        qs[4 * i + 0] = pi_e
        qs[4 * i + 1] = pi_x
        qs[4 * i + 2] = pi_y
        qs[4 * i + 3] = pi_z


    qs[4 * num_fsp] = pow(M_PI / 2.0, num_fsp - 1.0) \
        * pow(cme, 2.0 * num_fsp - 4.0) \
        / tgamma(num_fsp) \
        / tgamma(num_fsp - 1) \
        * pow(2.0 * M_PI, 4.0 - 3.0 * num_fsp)

    return qs


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef vector[double] __generate_ks(vector[double] masses, double cme, int num_fsp, vector[double] ps):
    """
    Generates a list of four-momentum with correct masses from massless
    four-momenta.

    Parameters
    ----------
    masses : double[:]
        List of masses of the final state particles.
    cme : double
        Center of mass energy of the process.
    ps : double[:]
        List of the massless four-momenta.

    Returns
    -------
    ks : double[:]
        List of the massive four-momenta with correct masses.
    """
    cdef int i
    cdef double xi
    cdef double k_e, k_x, k_y, k_z

    cdef double term1 = 0.0
    cdef double term2 = 0.0
    cdef double term3 = 1.0
    cdef double modulus

    xi = __find_root(masses, cme, num_fsp, ps)

    for i in range(num_fsp):
        k_e = sqrt(pow(masses[i], 2.0) + pow(xi * ps[4 * i + 0], 2.0))
        k_x = xi * ps[4 * i + 1]
        k_y = xi * ps[4 * i + 2]
        k_z = xi * ps[4 * i + 3]

        ps[4 * i + 0] = k_e
        ps[4 * i + 1] = k_x
        ps[4 * i + 2] = k_y
        ps[4 * i + 3] = k_z

        modulus = sqrt(pow(k_x, 2.0) + pow(k_y, 2.0) + pow(k_z, 2.0))

        term1 += modulus / cme
        term2 += pow(modulus, 2.0) / ps[4 * i + 0]
        term3 = term3 * modulus / ps[4 * i + 0]

    term1 = pow(term1, 2.0 * num_fsp - 3.0)
    term2 = pow(term2, -1.0)

    ps[4 * num_fsp] = ps[4 * num_fsp] * term1 * term2 * term3 * cme

    return ps


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def generate_point(vector[double] masses, double cme, int num_fsp):
    """
    Generate a single relativistic phase space point.

    Parameters
    ----------
    masses : double[:]
        List of masses of the final state particles.
    cme : double
        Center of mass energy of the process.
    rands : double[:]
        List of random numbers.

    Returns
    -------
    phase_space_point : double[:]
        List of four momenta and a event weight. The returned numpy array is of
        the form {ke1, kx1, ky1, kz1, ..., keN, kxN, kyN, kzN, weight}.
    """
    global rng
    cdef random_device rd
    rng = mt19937(rd())
    return c_generate_point(masses, cme, num_fsp)



@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef vector[double] c_generate_point(vector[double] masses, double cme, int num_fsp):
    """
    c version of generate_point.
    """

    return __generate_ks(masses, cme, num_fsp,
                         __generate_ps(masses, cme, num_fsp,
                                        __generate_qs(masses, cme, num_fsp)))


@cython.boundscheck(False)
@cython.wraparound(False)
def generate_space(int num_ps_pts, vector[double] masses, double cme, int num_fsp):
    """
    Generate a specified number of phase space points given a set of
    final state particles and a given center of mass energy.

    Parameters
    ----------
    num_ps_pts : int
        Total number of phase space points to generate.
    masses : numpy.ndarray
        List of masses of the final state particles.
    cme : double
        Center-of-mass-energy of the process.
    mat_elem_sqrd : (double)(numpy.ndarray) {lambda klist: 1}
        Function for the matrix element squared.

    Returns
    -------
    phase_space_points : numpy.ndarray
        List of phase space points. The phase space points are in the form
        {{ke11, kx11, ky11, kz11, ..., keN1, kxN1, kyN1, kzN1, weight1},
            .
            .
            .
         {ke1N, kx1N, ky1N, kz1N, ..., keNN, kxNN, kyNN, kzNN, weightN}}
    """
    return c_generate_space(num_ps_pts, masses, cme, num_fsp)



@cython.boundscheck(False)
@cython.wraparound(False)
cdef vector[vector[double]] c_generate_space(int num_ps_pts,
                                             vector[double] masses,
                                             double cme, int num_fsp):
    """
    Generate a specified number of phase space points given a set of
    final state particles and a given center of mass energy.

    Parameters
    ----------
    num_ps_pts : int
        Total number of phase space points to generate.
    masses : numpy.ndarray
        List of masses of the final state particles.
    cme : double
        Center-of-mass-energy of the process.
    mat_elem_sqrd : (double)(numpy.ndarray) {lambda klist: 1}
        Function for the matrix element squared.

    Returns
    -------
    phase_space_points : numpy.ndarray
        List of phase space points. The phase space points are in the form
        {{ke11, kx11, ky11, kz11, ..., keN1, kxN1, kyN1, kzN1, weight1},
            .
            .
            .
         {ke1N, kx1N, ky1N, kz1N, ..., keNN, kxNN, kyNN, kzNN, weightN}}
    """
    cdef int i

    global rng
    cdef random_device rd
    rng = mt19937(rd())

    cdef vector[vector[double]] space

    for i in range(num_ps_pts):
        space.push_back(c_generate_point(masses, cme, num_fsp))


    return space



"""
Module containing common field theory functions.

@author : Logan Morrison and Adam Coogan
@date : January 2018
"""

import numpy as np
cimport numpy as np
import cython

cdef np.ndarray metric_diag = np.array([1.0, -1.0, -1.0, -1.0])


@cython.boundscheck(False)
@cython.wraparound(False)
def minkowski_dot(np.ndarray[double, ndim=1] fv1,
                  np.ndarray[double, ndim=1] fv2):
    """
    Returns the dot product of two four vectors using the west coast metric.

    Paramaters
    ----------
    fv1 : numpy.ndarray[double, ndim=1]
        First four vector.
    fv2 : numpy.ndarray[double, ndim=1]
        Second four vector.

    Returns
    -------
    dot_product : double
        Returns fv1 * fv2.
    """
    return np.sum(metric_diag[:] * fv1[:] * fv2[:])



@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def cross_section_prefactor(double m1, double m2, double cme):
    """
    Returns the prefactor of the 2-N relativistic cross section.

    Paramaters
    ----------
    m1 : double
        Mass of particle one.
    m2 : double
        Mass of particle two.
    cme : double
        Center of mass energy.

    Returns
    -------
    prefactor : double
        Returns 1 / ((2 E1) (2 E2) |v1 - v2|)
    """
    cdef double E1 = (cme**2 + m1**2 - m2**2) / (2. * cme)
    cdef double E2 = (cme**2 + m2**2 - m1**2) / (2. * cme)

    cdef double p = np.sqrt((m1 - m2 - cme) * (m1 + m2 - cme) *
                            (m1 - m2 + cme) * (m1 + m2 + cme)) / (2. * cme)

    cdef double v1 = p / E1
    cdef double v2 = p / E2

    cdef double vrel = v1 + v2

    return 1.0 / (2.0 * E1) / (2.0 * E2) / vrel



import numpy as np

from hazma.decay import muon

from hazma.parameters import muon_mass as mmu
from hazma.parameters import electron_mass as me
from hazma.parameters import charged_pion_mass as mpi
from hazma.parameters import neutral_pion_mass as mpi0

from hazma.gamma_ray import gamma_ray_decay

# Stuff needed to compute fsr from x xbar -> P -> pip pim pi0
# from ..gamma_ray_decay import gamma_ray_fsr
# from .pseudo_scalar_mediator_mat_elem_sqrd_rambo import msqrd_xx_to_p_to_pm0g


class PseudoScalarMediatorSpectra:
    # TODO: pp spectrum. Gonna need Logan to do this since it
    # requires cython...
    def dnde_pp(self, egams, Q, mode="total"):
        # eng_p = Q / 2.
        pass

    def dnde_ee(self, egams, cme, spectrum_type="all"):
        """Computes spectrum from DM annihilation into electrons.
        """
        if spectrum_type == "all":
            return self.dnde_ee(egams, cme, "fsr") + self.dnde_ee(egams, cme, "decay")
        elif spectrum_type == "fsr":
            return self.dnde_xx_to_p_to_ffg(egams, cme, me)
        elif spectrum_type == "decay":
            return np.array([0.0 for _ in range(len(egams))])
        else:
            raise ValueError(
                "Type {} is invalid. Use 'all', 'fsr' or \
                             'decay'".format(
                    spectrum_type
                )
            )

    def dnde_mumu(self, egams, cme, spectrum_type="all"):
        """Computes spectrum from DM annihilation into muons.
        """
        if spectrum_type == "all":
            return self.dnde_mumu(egams, cme, "fsr") + self.dnde_mumu(
                egams, cme, "decay"
            )
        elif spectrum_type == "fsr":
            return self.dnde_xx_to_p_to_ffg(egams, cme, mmu)
        elif spectrum_type == "decay":
            return 2.0 * muon(egams, cme / 2.0)
        else:
            raise ValueError(
                "Type {} is invalid. Use 'all', 'fsr' or \
                             'decay'".format(
                    spectrum_type
                )
            )

    def dnde_pi0pipi(self, egams, cme, spectrum_type="all"):
        """Computes spectrum from DM annihilation into a neutral pion and two
        charged pions.

        Notes
        -----
        This function uses RAMBO to "convolve" the pions' spectra with the
        matrix element over the pi0 pi pi phase space.
        """
        if cme < 2.0 * mpi + mpi0:
            return np.array([0.0 for _ in range(len(egams))])

        if spectrum_type == "all":
            return self.dnde_pi0pipi(egams, cme, "fsr") + self.dnde_pi0pipi(
                egams, cme, "decay"
            )
        elif spectrum_type == "fsr":
            # Define the tree level and radiative matrix element squared for
            # RAMBO. These need to be of the form double(*func)(np.ndarray)
            # where
            # the np.ndarray is a list of 4-momenta. Note msqrd_xx_to_p_to_pm0
            # takes params as the second argument. The first and second FS
            # particles must be the charged pions and the third a neutral pion.

            # NOTE: I am removing this because it takes too long and need to
            # be extrapolated and evaluated at the correct egams

            """
            def msqrd_tree(momenta):
                return msqrd_xx_to_p_to_pm0(momenta)

            def msqrd_rad(momenta):
                return msqrd_xx_to_p_to_pm0g(momenta)

            isp_masses = np.array([params.mx, params.mx])
            fsp_masses = np.array([mpi, mpi, mpi0, 0.0])

            return gamma_ray_fsr(isp_masses, fsp_masses, cme,
                                   num_ps_pts=50000, num_bins=150,
                                   mat_elem_sqrd_tree=msqrd_tree,
                                   mat_elem_sqrd_rad=msqrd_rad)
            """

            return np.array([0.0 for _ in range(len(egams))])
        elif spectrum_type == "decay":
            # Define the matrix element squared for RAMBO. This needs to be
            # of the form double(*func)(np.ndarray) where the np.ndarray is
            # a list of 4-momenta. Note msqrd_xx_to_p_to_pm0 takes params as
            # the second argument. The first and second FS particles must be
            # the charged pions and the third a neutral pion.

            return gamma_ray_decay(
                ["charged_pion", "charged_pion", "neutral_pion"],
                cme,
                egams,
                num_ps_pts=1000,
                mat_elem_sqrd=self.msqrd_xx_to_p_to_pm0,
            )
        else:
            raise ValueError(
                "Type {} is invalid. Use 'all', 'fsr' or \
                             'decay'".format(
                    spectrum_type
                )
            )

    def dnde_pi0pi0pi0(self, egams, cme, spectrum_type="all"):
        """Return the gamma ray spectrum for dark matter annihilations into
        three neutral pions.

        Notes
        -----
        This function uses RAMBO to "convolve" the pions' spectra with the
        matrix element over the pi0 pi0 pi0 phase space.
        """
        if cme < 3.0 * mpi0:
            return np.array([0.0 for _ in range(len(egams))])

        if spectrum_type == "all":
            return self.dnde_pi0pi0pi0(egams, cme, "decay")
        elif spectrum_type == "fsr":
            return np.array([0.0 for _ in range(len(egams))])
        elif spectrum_type == "decay":
            # Define the matrix element squared for RAMBO. This needs to be
            # of the form double(*func)(np.ndarray) where the np.ndarray is
            # a list of 4-momenta. Note msqrd_xx_to_p_to_000 takes params as
            # the second argument.
            def msqrd_tree(momenta):
                return self.msqrd_xx_to_p_to_000(momenta)

            return gamma_ray_decay(
                ["neutral_pion", "neutral_pion", "neutral_pion"],
                cme,
                egams,
                num_ps_pts=1000,
                mat_elem_sqrd=msqrd_tree,
            )
        else:
            raise ValueError(
                "Type {} is invalid. Use 'all', 'fsr' or \
                             'decay'".format(
                    spectrum_type
                )
            )

    def spectrum_funcs(self):
        """
        Returns a dictionary of all the avaiable spectrum functions for
        a pair of initial state fermions with mass `mx` annihilating into
        each available final state.

        Each argument of the spectrum functions in `eng_gams`, an array
        of the gamma ray energies to evaluate the spectra at and `cme`, the
        center of mass energy of the process.
        """
        return {
            "mu mu": self.dnde_mumu,
            "e e": self.dnde_ee,
            "pi0 pi pi": self.dnde_pi0pipi,
            "pi0 pi0 pi0": self.dnde_pi0pi0pi0,
            "p p": self.dnde_pp,
        }

    def gamma_ray_lines(self, cme):
        bf = self.annihilation_branching_fractions(cme)["g g"]

        return {"g g": {"energy": cme / 2.0, "bf": bf}}



"""
Module for integrating over thee body phase space.

@author : Logan Morrison and Adam Coogan
@data : January 2018
"""

import numpy as np
cimport numpy as np
import cython
from scipy.integrate import quad


@cython.cdivision(True)
def s_to_E1(double s, double m1, double M):
    """
    Returns the energy of particle 1 using the mandelstam variable s.

    Parameters
    ----------
    s : double
        Mandelstam variable associated with m1, defined as (P-p1)^2, where P = p1 + p2 + p3.
    m1 : double
        Mass of particle 1.
    M : double
        Center of mass energy, or sqrt((p1 + p2 + p3)^2).

    Returns
    -------
    E1 : double
        Energy of particle 1.
    """
    return (M**2 + m1**2 - s) / (2 * M)


@cython.cdivision(True)
def E1_to_s(E1, m1, M):
    """
    Returns the mandelstam variable s of particle 1 using the energy.

    Parameters
    ----------
    E! : double
        Energy of particle 1.
    m1 : double
        Mass of particle 1.
    M : double
        Center of mass energy, or sqrt((p1 + p2 + p3)^2).

    Returns
    -------
    s : double
        Mandelstam variable associated with m1, defined as (P-p1)^2, where P = p1 + p2 + p3.
    """
    return M**2 - 2 * M * E1 + m1**2


@cython.cdivision(True)
def t_to_E2(double t, double m2, double M):
    """
    Returns the energy of particle 2 using the mandelstam variable t.

    Parameters
    ----------
    t : double
        Mandelstam variable associated with m2, defined as (P-p2)^2, where P = p1 + p2 + p3.
    m2 : double
        Mass of particle 2.
    M : double
        Center of mass energy, or sqrt((p1 + p2 + p3)^2).

    Returns
    -------
    E2 : double
        Energy of particle 2.
    """
    return (M**2 + m2**2 - t) / (2 * M)


@cython.cdivision(True)
def st_to_u(double m1, double m2, double m3, double M, double s, double t):
    """
    Returns the value of the mandelstam variable u using the mandelstam variables s and t.

    Parameters
    ----------
    m1 : double
        Mass of particle 1.
    m2 : double
        Mass of particle 2.
    m3 : double
        Mass of particle 3.
    M : double
        Center of mass energy, or sqrt((p1 + p2 + p3)^2).
    s : double
        Mandelstam variable associated with m1, defined as (P-p1)^2, where P = p1 + p2 + p3.
    t : double
        Mandelstam variable associated with m2, defined as (P-p2)^2, where P = p1 + p2 + p3.

    Returns
    -------
    u : double
        Mandelstam variable associated with m3, defined as (P-p3)^2, where P = p1 + p2 + p3.
    """
    return M**2 + m1**2 + m2**2 + m3**2 - s - t


@cython.cdivision(True)
def phase_space_prefactor(double M):
    r"""
    Returns the normalization of the differential phase space volume is terms
    of s and t

    Parameters
    ----------
    M : double
        Center of mass energy, or sqrt((p1 + p2 + p3)^2).

    Returns
    -------
    prefactor : double
        The prefactor of ds*dt. That is, the factor, :math:`C` that equates
        :math:`d\phi_3 = C dt st`

    """
    return 1.0 / (16.0 * M**2 * (2. * np.pi)**3)


@cython.cdivision(True)
def t_lim1(double s, double m1, double m2, double m3, double M):
    """
    Returns the value of the first boundary curve evaluated at the mandelstam
    variable s.

    Parameters
    ----------
    s : double
        Mandelstam variable associated with m1, defined as (P-p1)^2, where P = p1 + p2 + p3.
    m1 : double
        Mass of particle 1.
    m2 : double
        Mass of particle 2.
    m3 : double
        Mass of particle 3.
    M : double
        Center of mass energy, or sqrt((p1 + p2 + p3)^2).

    Returns
    -------
    The first limiting curve of the three-body phase space.
    """
    return -((M - m1) * (M + m1) * (m2 - m3) * (m2 + m3) -
             (M**2 + m1**2 + m2**2 + m3**2) * s + s**2 +
             np.sqrt(M**4 + (m1**2 - s)**2 - 2 * M**2 *
                  (m1**2 + s)) * np.sqrt(m2**4 + (m3**2 - s)**2 - 2 * m2**2 *
                                      (m3**2 + s))) / (2. * s)


@cython.cdivision(True)
def t_lim2(double s, double m1, double m2, double m3, double M):
    """
    Returns the value of the second boundary curve evaluated at the mandelstam
    variable s.

    Parameters
    ----------
    s : double
        Mandelstam variable associated with m1, defined as (P-p1)^2, where P = p1 + p2 + p3.
    m1 : double
        Mass of particle 1.
    m2 : double
        Mass of particle 2.
    m3 : double
        Mass of particle 3.
    M : double
        Center of mass energy, or sqrt((p1 + p2 + p3)^2).

    Returns
    -------
    The seconds limiting curve of the three-body phase space.
    """
    return ((-M + m1) * (M + m1) * (m2 - m3) * (m2 + m3) +
            (M**2 + m1**2 + m2**2 + m3**2) * s - s**2 +
            np.sqrt(M**4 + (m1**2 - s)**2 - 2 * M**2 * (m1**2 + s)) * np.sqrt(
                m2**4 + (m3**2 - s)**2 - 2 * m2**2 * (m3**2 + s))) / (2. * s)


@cython.cdivision(True)
def s_max(double m1, double m2, double m3, double M):
    """
    Returns the maximum value of the mandelstam variable s.

    Parameters
    ----------
    m1 : double
        Mass of particle 1.
    m2 : double
        Mass of particle 2.
    m3 : double
        Mass of particle 3.
    M : double
        Center of mass energy, or sqrt((p1 + p2 + p3)^2).

    Returns
    -------
    The maximum value that s = M - 2 E1*m1 + m1^2 can have.
    """
    return (M - m1)**2


@cython.cdivision(True)
def s_min(double m1, double m2, double m3, double M):
    """
    Returns the minimum value of the mandelstam variable s.

    Parameters
    ----------
    m1 : double
        Mass of particle 1.
    m2 : double
        Mass of particle 2.
    m3 : double
        Mass of particle 3.
    M : double
        Center of mass energy, or sqrt((p1 + p2 + p3)^2).

    Returns
    -------
    The minimum value that s = M - 2 E1*m1 + m1^2 can have.
    """
    return (m2 + m3)**2


@cython.cdivision(True)
def t_integral(double s, double m1, double m2, double m3, double M,
               mat_elem_sqrd):
    """
    Returns the integral over the mandelstam variable t.

    Parameters
    ----------
    s : double
        Mandelstam variable associated with m1, defined as (P-p1)^2, where P = p1 + p2 + p3.
    m1 : double
        Mass of particle 1.
    m2 : double
        Mass of particle 2.
    m3 : double
        Mass of particle 3.
    M : double
        Center of mass energy, or sqrt((p1 + p2 + p3)^2).
    mat_elem_sqrd : double(*func)(double t)
        Function for the matrix element squared. Argument is the mandelstam variable t associated with m2, defined as (P-p2)^2, where P = p1 + p2 + p3.

    Returns
    -------
    t_int : double
        Value of the integral over mandelstam varible t given the mandelstam varible s.
    """
    cdef double t_1 = t_lim1(s, m1, m2, m3, M)
    cdef double t_2 = t_lim2(s, m1, m2, m3, M)

    cdef double t_max = 0.0
    cdef double t_min = 0.0

    if t_1 > t_2:
        t_max = t_1
        t_min = t_2
    else:
        t_max = t_2
        t_min = t_1

    return quad(mat_elem_sqrd, t_min, t_max)


@cython.cdivision(True)
def three_body_phase_space_integral(double m1, double m2, double m3, double M,
                                    mat_elem_sqrd):
    """
    Returns the integral over the mandelstam variable s.

    Parameters
    ----------
    m1 : double
        Mass of particle 1.
    m2 : double
        Mass of particle 2.
    m3 : double
        Mass of particle 3.
    M : double
        Center of mass energy, or sqrt((p1 + p2 + p3)^2).
    mat_elem_sqrd : double(*func)(double s, double t)
        Function for the matrix element squared. Argument is the mandelstam variable s associated with m1, defined as (P-p1)^2, where P = p1 + p2 + p3.

    Returns
    -------
    t_int : double
        Value of the integral over mandelstam varible s.
    """
    cdef double smax = s_max(m1, m2, m3, M)
    cdef double smin = s_min(m1, m2, m3, M)

    def integrand(s):
        """
        Integrand for s integral
        """
        def mat_elem_sqrd_t(t):
            """
            Integrand for t integral
            """
            return mat_elem_sqrd(s, t)

        return t_integral(s, m1, m2, m3, M, mat_elem_sqrd_t)[0]

    return quad(integrand, smin, smax)



import numpy as np
cimport numpy as np
from scipy.integrate import quad
from libc.math cimport exp, log, M_PI, log10, sqrt
import cython
from functools import partial
include "parameters.pxd"
import warnings


"""
Module for computing the photon spectrum from radiative muon decay.

The radiative spectrum of the muon was taken from: arXiv:hep-ph/9909265
"Muon Decay and Physics Beyond the Standard Model".
"""

@cython.cdivision(True)
cdef double __j_plus(double y):
    """
    Form factor in differential branching fraction of radiative muon decay.
    See p.18, eqn (54) of arXiv:hep-ph/9909265.

    Keyword arguments::
        y -- 2 * (photon energy) / (muon mass)
    """
    cdef double yconj = 1.0 - y
    cdef double r = RATIO_E_MU_MASS_SQ
    cdef double preFactor = ALPHA_EM * yconj / 6.0 / M_PI
    cdef double term1 = 3.0 * log(yconj / r) - (17.0 / 2.0)
    cdef double term2 = -3.0 * log(yconj / r) + 7.0
    cdef double term3 = 2.0 * log(yconj / r) - (13.0 / 3.0)
    return preFactor * (term1 + term2 * yconj + term3 * yconj**2.0)

@cython.cdivision(True)
cdef double __j_minus(double y):
    """
    Form factor in differential branching fraction of radiative muon decay.
    See p.18, eqn (55) of arXiv:hep-ph/9909265.

    Keyword arguments::
        y -- 2 * (photon energy) / (muon mass)
    """
    cdef double yconj = 1.0 - y
    cdef double r = RATIO_E_MU_MASS_SQ
    cdef double preFactor = ALPHA_EM * yconj**2.0 / 6.0 / M_PI
    cdef double term1 = 3.0 * log(yconj / r) - (93.0 / 12.0)
    cdef double term2 = -4.0 * log(yconj / r) + (29.0 / 3.0)
    cdef double term3 = 2.0 * log(yconj / r) - (55.0 / 12.0)
    return preFactor * (term1 + term2 * yconj + term3 * yconj**2.0)

@cython.cdivision(True)
cdef double __dBdy(double y):
    """
    Differential branching fraction from: mu -> e nu nu gam.
    See p.18, eqn (56) of arXiv:hep-ph/9909265.

    Keyword arguments::
        y -- 2 * (photon energy) / (muon mass)
    """
    cdef double result = 0.0
    if 0.0 <= y and y <= 1.0 - RATIO_E_MU_MASS_SQ:
        result = (2.0 / y) * (__j_plus(y) + __j_minus(y))
    return result

@cython.cdivision(True)
cdef double __gamma(double eng, double mass):
    """
    Returns special relativity boost factor gamma.

    Keyword arguments::
        eng -- energy of particle.
        mass -- mass of particle.
    """
    return eng / mass

@cython.cdivision(True)
cdef double __beta(double eng, double mass):
    """
    Returns velocity in natural units.

    Keyword arguments::
        eng -- energy of particle.
        mass -- mass of particle.
    """
    return sqrt(1.0 - (mass / eng)**2.0)

@cython.cdivision(True)
cdef double __integrand(double cl, double engGam, double engMu):
    """
    Compute integrand of dN_{\gamma}/dE_{\gamma} from mu -> e nu nu gamma
    in laboratory frame. The integration variable is cl - the angle between
    gamma ray and muon.

    Keyword arguments::
        cl -- Angle between gamma ray and muon in laboratory frame.
        engGam -- Gamma ray energy in laboratory frame.
        engMu -- Muon energy in laboratory frame.
    """
    cdef double beta = __beta(engMu, MASS_MU)
    cdef double gamma = __gamma(engMu, MASS_MU)

    cdef double engGamMuRF = gamma * engGam * (1.0 - beta * cl)

    return __dBdy((2.0 / MASS_MU) * engGamMuRF) \
        / (engMu * (1.0 - cl * beta))


cdef double CSpectrumPoint(double eng_gam, double eng_mu):
    """
    Compute dN_{\gamma}/dE_{\gamma} from mu -> e nu nu gamma in the
    laborartory frame.

    Keyword arguments::
        eng_gam (float) -- Gamma ray energy in laboratory frame.
        eng_mu (float) -- Muon energy in laboratory frame.
    """
    message = 'Energy of kaon cannot be less than the kaon mass. Returning 0.'
    if eng_mu < MASS_MU:
        # raise warnings.warn(message, RuntimeWarning)
        return 0.0

    cdef double result = 0.0

    cdef double beta = __beta(eng_mu, MASS_MU)
    cdef double gamma = __gamma(eng_mu, MASS_MU)

    cdef double eng_gam_max = 0.5 * (MASS_MU - MASS_E**2.0 / MASS_MU) \
        * gamma * (1.0 + beta)

    if 0 <= eng_gam and eng_gam <= eng_gam_max:
        result = quad(__integrand, -1.0, 1.0, args=(eng_gam, eng_mu), \
            points=[-1.0, 1.0], epsabs=10**-10., epsrel=10**-4.)[0]


    return result


@cython.cdivision(True)
cdef np.ndarray CSpectrum(np.ndarray eng_gams, double eng_mu):
    """
    Compute dN/dE from mu -> e nu nu gamma in the laborartory frame.

    Paramaters
    ----------
    eng_gams : np.ndarray
        List of gamma ray energies in laboratory frame.
    eng_mu : float
        Muon energy in laboratory frame.

    Returns
    -------
    spec : np.ndarray
        List of gamma ray spectrum values, dNdE, evaluated at `eng_gams`
        given muon energy `eng_mu`.
    """
    cdef int numpts = len(eng_gams)

    cdef np.ndarray spec = np.zeros(numpts, dtype=np.float64)

    for i in range(numpts):
        spec[i] = CSpectrumPoint(eng_gams[i], eng_mu)

    return spec

def SpectrumPoint(double eng_gam, double eng_mu):
    """
    Compute dN_{\gamma}/dE_{\gamma} from mu -> e nu nu gamma in the
    laborartory frame.

    Keyword arguments::
        eng_gam (float) -- Gamma ray energy in laboratory frame.
        eng_mu (float) -- Muon energy in laboratory frame.
    """
    return CSpectrumPoint(eng_gam, eng_mu)



@cython.cdivision(True)
def Spectrum(np.ndarray eng_gams, double eng_mu):
    """
    Compute dN/dE from mu -> e nu nu gamma in the laborartory frame.

    Paramaters
    ----------
    eng_gams : np.ndarray
        List of gamma ray energies in laboratory frame.
    eng_mu : float
        Muon energy in laboratory frame.

    Returns
    -------
    spec : np.ndarray
        List of gamma ray spectrum values, dNdE, evaluated at `eng_gams`
        given muon energy `eng_mu`.
    """
    return CSpectrum(eng_gams, eng_mu)



"""Module containing squared matrix elements from axial-vector mediator models.

@author - Logan Morrison and Adam Coogan
@date - December 2017
"""
import numpy as np
from hazma.field_theory_helper_functions.common_functions import minkowski_dot
from hazma.parameters import alpha_em


def msqrd_xx_to_a_to_ff(moms, mx, mf, ma, cxxa, cffa):
    """Returns the spin-averaged, squared matrix element for a pair of fermions,
    *x*, annihilating into a pair of fermions, *f*, through a axial-vector
    mediator in the s-channel.

    Parameters
    ----------
    moms : numpy.ndarray
        Array of four momenta of the final state particles. The first two must
        for momenta must be the fermions and the last must be the photon.
        moms must be in the form {{ke1, kx1, ky1, kz1}, ..., {keN, kxN, kyN,
        kzN}}.
    mx : float
        Mass of incoming fermions.
    mf : float
        Mass of final state fermions.
    ma : float
        Mass of axial-vector mediator.
    qf : float
        Electric charge of final state fermion.
    cxxa : float
        Coupling of initial state fermions with the axial-vector mediator.
    cffa : float
        Coupling of final state fermions with the axial-vector mediator.

    Returns
    -------
    mat_elem_sqrd : float
        Spin averaged, squared matrix element for x + x -> A^* -> f + f.
    """
    p3 = moms[0]
    p4 = moms[1]

    Q = p3[0] + p4[0]

    E = Q / 2.0
    p = np.sqrt(E ** 2 - mx ** 2)

    p1 = np.array([E, 0.0, 0.0, p])
    p2 = np.array([E, 0.0, 0.0, -p])

    p1DOTp4 = minkowski_dot(p1, p4)
    p2DOTp3 = minkowski_dot(p2, p3)
    p1DOTp3 = minkowski_dot(p1, p3)
    p2DOTp4 = minkowski_dot(p2, p4)

    return (
        4
        * cffa ** 2
        * cxxa ** 2
        * (
            2 * mf ** 2 * mx ** 2
            + 6 * mf ** 2 * mx ** 2
            + 2 * p1DOTp4 * p2DOTp3
            + 2 * p1DOTp3 * p2DOTp4
            - mf ** 2 * Q ** 2
            - mx ** 2 * Q ** 2
        )
    ) / (-ma ** 2 + Q ** 2) ** 2


def msqrd_xx_to_a_to_ffg(moms, mx, mf, ma, qf, cxxa, cffa):
    """Returns the spin-averaged, squared matrix element for a pair of fermions,
    *x*, annihilating into a pair of fermions, *f*, and a photon through a
    axial-vector mediator in the s-channel.

    Parameters
    ----------
    moms : numpy.ndarray
        Array of four momenta of the final state particles. The first two must
        for momenta must be the fermions and the last must be the photon.
        moms must be in the form {{ke1, kx1, ky1, kz1}, ..., {keN, kxN, kyN,
        kzN}}.
    mx : float
        Mass of incoming fermions.
    mf : float
        Mass of final state fermions.
    ma : float
        Mass of axial-vector mediator.
    qf : float
        Electric charge of final state fermion.
    cxxa : float
        Coupling of initial state fermions with the axial-vector mediator.
    cffa : float
        Coupling of final state fermions with the axial-vector mediator.

    Returns
    -------
    mat_elem_sqrd : float
        Spin averaged, squared matrix element for x + x -> a^* -> f + f + g.
    """
    p3 = moms[0]
    p4 = moms[1]
    k = moms[2]

    Q = p3[0] + p4[0] + k[0]

    E = Q / 2
    p = np.sqrt(E ** 2 - mx ** 2)

    p1 = np.array([E, 0, 0, p])
    p2 = np.array([E, 0, 0, -p])

    kDOTp3 = minkowski_dot(k, p3)
    kDOTp4 = minkowski_dot(k, p4)
    p3DOTp4 = minkowski_dot(p3, p4)

    kDOTp1 = minkowski_dot(k, p1)
    p1DOTp3 = minkowski_dot(p1, p3)
    p1DOTp4 = minkowski_dot(p1, p4)

    kDOTp2 = minkowski_dot(k, p2)
    p2DOTp3 = minkowski_dot(p2, p3)
    p2DOTp4 = minkowski_dot(p2, p4)

    e = np.sqrt(4 * np.pi * alpha_em)

    return (
        4
        * cffa ** 2
        * cxxa ** 2
        * e ** 2
        * (
            2
            * kDOTp3
            * kDOTp4
            * (
                kDOTp1 * kDOTp3 * p2DOTp3
                - 2 * kDOTp4 * p1DOTp3 * p2DOTp3
                + kDOTp3 * p1DOTp4 * p2DOTp3
                + kDOTp4 * p1DOTp4 * p2DOTp3
                + kDOTp1 * kDOTp4 * p2DOTp4
                + kDOTp3 * p1DOTp3 * p2DOTp4
                + kDOTp4 * p1DOTp3 * p2DOTp4
                - 2 * kDOTp3 * p1DOTp4 * p2DOTp4
                + ((kDOTp1 + 2 * p1DOTp4) * p2DOTp3 + (kDOTp1 + 2 * p1DOTp3) * p2DOTp4)
                * p3DOTp4
                + kDOTp2
                * (kDOTp3 * p1DOTp3 + kDOTp4 * p1DOTp4 + (p1DOTp3 + p1DOTp4) * p3DOTp4)
                - mx ** 2
                * (
                    kDOTp3 ** 2
                    + kDOTp4 ** 2
                    + 2 * (kDOTp3 + kDOTp4) * p3DOTp4
                    + 2 * p3DOTp4 ** 2
                )
            )
            + (kDOTp3 ** 2 + kDOTp4 ** 2) * mf ** 4 * (-6 * mx ** 2 + Q ** 2)
            + 2
            * mf ** 2
            * (
                -(kDOTp2 * kDOTp3 ** 2 * p1DOTp3)
                - kDOTp2 * kDOTp4 ** 2 * p1DOTp4
                - kDOTp3 ** 2 * p1DOTp4 * p2DOTp3
                - kDOTp4 ** 2 * p1DOTp4 * p2DOTp3
                - kDOTp3 ** 2 * p1DOTp3 * p2DOTp4
                - kDOTp4 ** 2 * p1DOTp3 * p2DOTp4
                + kDOTp1
                * (
                    2 * kDOTp2 * kDOTp3 * kDOTp4
                    - kDOTp3 ** 2 * p2DOTp3
                    - kDOTp4 ** 2 * p2DOTp4
                )
                + mx ** 2
                * (
                    kDOTp3 ** 3
                    + kDOTp3 ** 2 * (kDOTp4 + p3DOTp4)
                    + kDOTp4 ** 2 * (kDOTp4 + p3DOTp4)
                    + kDOTp3 * kDOTp4 * (kDOTp4 + 6 * p3DOTp4)
                )
                - kDOTp3 * kDOTp4 * p3DOTp4 * Q ** 2
            )
        )
        * qf ** 2
    ) / (
        kDOTp3 ** 2
        * kDOTp4 ** 2
        * (ma ** 2 - 2 * mf ** 2 - 2 * (kDOTp3 + kDOTp4 + p3DOTp4)) ** 2
    )



from cmath import sqrt, pi, atanh

from hazma.parameters import vh, b0, alpha_em
from hazma.parameters import charged_pion_mass as mpi
from hazma.parameters import neutral_pion_mass as mpi0
from hazma.parameters import up_quark_mass as muq
from hazma.parameters import down_quark_mass as mdq
from hazma.parameters import muon_mass as mmu
from hazma.parameters import electron_mass as me

from scipy.integrate import quad


class ScalarMediatorCrossSection:
    def sigma_xx_to_s_to_ff(self, e_cm, f):
        """Returns the spin-averaged, cross section for dark matter
        annihilating into a pair of fermions, *f* through a scalar mediator in
        the s-channel.

        Parameters
        ----------
        e_cm : float
            Center of mass energy.
        f: str
            String for the final state fermion: f = 'e' for electron and
            f = 'mu' for muon.

        Returns
        -------
        sigma : float
            Cross section for x + x -> s* -> f + f.
        """
        mx = self.mx

        if f == "e":
            mf = me
        elif f == "mu":
            mf = mmu

        if e_cm > 2.0 * mf and e_cm >= 2.0 * mx:
            ms = self.ms
            gsff = self.gsff
            gsxx = self.gsxx
            width_s = self.width_s

            ret_val = (
                gsff ** 2
                * gsxx ** 2
                * mf ** 2
                * (-2 * mx + e_cm)
                * (2 * mx + e_cm)
                * (-4 * mf ** 2 + e_cm ** 2) ** 1.5
            ) / (
                16.0
                * pi
                * e_cm ** 2
                * sqrt(-4 * mx ** 2 + e_cm ** 2)
                * vh ** 2
                * (
                    ms ** 4
                    - 2 * ms ** 2 * e_cm ** 2
                    + e_cm ** 4
                    + ms ** 2 * width_s ** 2
                )
            )

            assert ret_val.imag == 0
            assert ret_val.real >= 0

            return ret_val.real
        else:
            return 0.0

    def sigma_xx_to_s_to_gg(self, e_cm):
        """Returns the spin-averaged, cross section for dark matter
        annihilating into a pair of photons through a scalar mediator in the
        s-channel.

        Parameters
        ----------
        e_cm : float
            Center of mass energy.

        Returns
        -------
        sigma : float
            Cross section for x + x -> s* -> g + g.
        """
        mx = self.mx

        if e_cm >= 2.0 * mx:
            gsFF = self.gsFF
            gsxx = self.gsxx
            ms = self.ms
            widths = self.width_s
            rx = mx / e_cm
            Lam = self.lam

            ret_val = (
                alpha_em ** 2
                * gsFF ** 2
                * gsxx ** 2
                * e_cm ** 4
                * sqrt(1 - 4 * rx ** 2)
            ) / (
                64.0
                * Lam ** 2
                * pi ** 3
                * (ms ** 4 + e_cm ** 4 + ms ** 2 * (-2 * e_cm ** 2 + widths ** 2))
            )

            assert ret_val.imag == 0
            assert ret_val.real >= 0

            return ret_val.real
        else:
            return 0.0

    def sigma_xx_to_s_to_pi0pi0(self, e_cm):
        """Returns the spin-averaged, cross section for dark matter
        annihilating into a pair of neutral pion through a scalar mediator in
        the s-channel.

        Parameters
        ----------
        e_cm : float
            Center of mass energy.

        Returns
        -------
        sigma : float
            Cross section for x + x -> s* -> pi0 + pi0.
        """
        mx = self.mx

        if e_cm > 2.0 * mpi0 and e_cm >= 2.0 * mx:
            gsxx = self.gsxx
            gsff = self.gsff
            gsGG = self.gsGG
            ms = self.ms
            vs = self.vs
            widths = self.width_s
            Lam = self.lam
            rpi0 = mpi0 / e_cm
            rx = mx / e_cm

            ret_val = (
                gsxx ** 2
                * sqrt((-1 + 4 * rpi0 ** 2) * (-1 + 4 * rx ** 2))
                * (
                    162 * gsGG * Lam ** 3 * e_cm ** 2 * (-1 + 2 * rpi0 ** 2) * vh ** 2
                    + b0
                    * (mdq + muq)
                    * (9 * Lam + 4 * gsGG * vs)
                    * (
                        27 * gsff ** 2 * Lam ** 2 * vs * (3 * Lam + 4 * gsGG * vs)
                        - 2
                        * gsGG
                        * vh ** 2
                        * (
                            27 * Lam ** 2
                            - 30 * gsGG * Lam * vs
                            + 8 * gsGG ** 2 * vs ** 2
                        )
                        + gsff
                        * (-81 * Lam ** 3 * vh + 48 * gsGG ** 2 * Lam * vh * vs ** 2)
                    )
                )
                ** 2
            ) / (
                209952.0
                * Lam ** 6
                * pi
                * vh ** 4
                * (9 * Lam + 4 * gsGG * vs) ** 2
                * (ms ** 4 + e_cm ** 4 + ms ** 2 * (-2 * e_cm ** 2 + widths ** 2))
            )

            assert ret_val.imag == 0
            assert ret_val.real >= 0

            return ret_val.real
        else:
            return 0.0

    def sigma_xx_to_s_to_pipi(self, e_cm):
        """Returns the spin-averaged, cross section for dark matter
        annihilating into a pair of charged pions through a scalar mediator in
        the s-channel.

        Parameters
        ----------
        e_cm : float
            Center of mass energy.

        Returns
        -------
        sigma : float
            Cross section for x + x -> s* -> pi + pi.
        """
        mx = self.mx

        if e_cm > 2.0 * mpi and e_cm >= 2.0 * mx:
            gsxx = self.gsxx
            gsff = self.gsff
            gsGG = self.gsGG
            ms = self.ms
            vs = self.vs
            widths = self.width_s
            Lam = self.lam
            rpi = mpi / e_cm
            rx = mx / e_cm

            ret_val = (
                gsxx ** 2
                * sqrt((-1 + 4 * rpi ** 2) * (-1 + 4 * rx ** 2))
                * (
                    162 * gsGG * Lam ** 3 * e_cm ** 2 * (-1 + 2 * rpi ** 2) * vh ** 2
                    + b0
                    * (mdq + muq)
                    * (9 * Lam + 4 * gsGG * vs)
                    * (
                        27 * gsff ** 2 * Lam ** 2 * vs * (3 * Lam + 4 * gsGG * vs)
                        - 2
                        * gsGG
                        * vh ** 2
                        * (
                            27 * Lam ** 2
                            - 30 * gsGG * Lam * vs
                            + 8 * gsGG ** 2 * vs ** 2
                        )
                        + gsff
                        * (-81 * Lam ** 3 * vh + 48 * gsGG ** 2 * Lam * vh * vs ** 2)
                    )
                )
                ** 2
            ) / (
                104976.0
                * Lam ** 6
                * pi
                * vh ** 4
                * (9 * Lam + 4 * gsGG * vs) ** 2
                * (ms ** 4 + e_cm ** 4 + ms ** 2 * (-2 * e_cm ** 2 + widths ** 2))
            )

            assert ret_val.imag == 0
            assert ret_val.real >= 0

            return ret_val.real
        else:
            return 0.0

    def sigma_xx_to_ss(self, e_cm):
        """Returns the spin-averaged, cross section for dark matter
        annihilating into a pair of scalar mediator through the t and u
        channels.

        Parameters
        ----------
        e_cm : float
            Center of mass energy.

        Returns
        -------
        sigma : float
            Cross section for x + x -> s + s.
        """
        ms = self.ms
        mx = self.mx

        if e_cm > 2.0 * ms and e_cm >= 2.0 * mx:
            gsxx = self.gsxx

            ret_val = -(
                gsxx ** 4
                * sqrt(-4 * ms ** 2 + e_cm ** 2)
                * sqrt(-4 * mx ** 2 + e_cm ** 2)
                * (
                    -2 / (4 * mx ** 2 - e_cm ** 2)
                    - (ms ** 2 - 4 * mx ** 2) ** 2
                    / (
                        (4 * mx ** 2 - e_cm ** 2)
                        * (ms ** 4 - 4 * ms ** 2 * mx ** 2 + mx ** 2 * e_cm ** 2)
                    )
                    - (
                        2
                        * (
                            6 * ms ** 4
                            - 32 * mx ** 4
                            + 16 * mx ** 2 * e_cm ** 2
                            + e_cm ** 4
                            - 4 * ms ** 2 * (4 * mx ** 2 + e_cm ** 2)
                        )
                        * atanh(
                            (
                                sqrt(-4 * ms ** 2 + e_cm ** 2)
                                * sqrt(-4 * mx ** 2 + e_cm ** 2)
                            )
                            / (-2 * ms ** 2 + e_cm ** 2)
                        )
                    )
                    / (
                        sqrt(-4 * ms ** 2 + e_cm ** 2)
                        * (-2 * ms ** 2 + e_cm ** 2)
                        * (-4 * mx ** 2 + e_cm ** 2) ** 1.5
                    )
                )
            ) / (16.0 * pi * e_cm ** 2)

            assert ret_val.imag == 0
            assert ret_val.real >= 0

            return ret_val.real
        else:
            return 0.0

    def sigma_xx_to_s_to_xx(self, e_cm):
        """Returns the spin-averaged, self interaction cross section for dark
        matter.

        Parameters
        ----------
        e_cm : float
            Center of mass energy.

        Returns
        -------
        sigma : float
            Cross section for x + x -> x + x
        """
        rs = self.ms / e_cm
        rx = self.mx / e_cm
        gsxx = self.gsxx
        rws = self.width_s / e_cm

        if e_cm > 2.0 * self.mx:

            def msqrd(z):
                return (
                    gsxx ** 4
                    * (
                        3 * (-1 + z) ** 2
                        + 256 * rx ** 8 * (-1 + z) ** 2
                        - 64 * rx ** 6 * (5 - 8 * z + 3 * z ** 2)
                        + 32 * rx ** 4 * (5 - 6 * z + 3 * z ** 2)
                        - 4 * rx ** 2 * (5 - 12 * z + 7 * z ** 2)
                        + rs ** 4
                        * (
                            3
                            + z ** 2
                            - 8 * rx ** 2 * (1 + z ** 2)
                            + 16 * rx ** 4 * (3 + z ** 2)
                        )
                        + rs ** 2
                        * (
                            3
                            - 3 * z ** 2
                            - 64 * rx ** 6 * (3 - 4 * z + z ** 2)
                            - 16 * rx ** 4 * (-9 + 12 * z + z ** 2)
                            + 4 * rx ** 2 * (-17 + 8 * z + 5 * z ** 2)
                            + rws ** 2
                            * (
                                3
                                + z ** 2
                                - 8 * rx ** 2 * (1 + z ** 2)
                                + 16 * rx ** 4 * (3 + z ** 2)
                            )
                        )
                    )
                ) / (
                    (1 + rs ** 4 + rs ** 2 * (-2 + rws ** 2))
                    * (
                        4 * rs ** 4
                        + 4 * rs ** 2 * (rws ** 2 + (-1 + 4 * rx ** 2) * (-1 + z))
                        + (1 - 4 * rx ** 2) ** 2 * (-1 + z) ** 2
                    )
                )

            ret_val = quad(msqrd, -1, 1)[0] / (32.0 * pi * e_cm)

            assert ret_val.imag == 0.0
            assert ret_val.real >= 0.0

            return ret_val.real
        else:
            return 0.0

    def annihilation_cross_section_funcs(self):
        return {
            "mu mu": lambda e_cm: self.sigma_xx_to_s_to_ff(e_cm, "mu"),
            "e e": lambda e_cm: self.sigma_xx_to_s_to_ff(e_cm, "e"),
            "g g": self.sigma_xx_to_s_to_gg,
            "pi0 pi0": self.sigma_xx_to_s_to_pi0pi0,
            "pi pi": self.sigma_xx_to_s_to_pipi,
            "s s": self.sigma_xx_to_ss,
        }



from hazma.field_theory_helper_functions.common_functions import minkowski_dot as MDot

from hazma.parameters import charged_pion_mass as mpi
from hazma.parameters import neutral_pion_mass as mpi0
from hazma.parameters import up_quark_mass as muq
from hazma.parameters import down_quark_mass as mdq
from hazma.parameters import b0, vh, fpi, qe

import numpy as np


class PseudoScalarMediatorMSqrdRambo:
    def msqrd_xx_to_p_to_pm0(self, momenta):
        """
        Returns the squared matrix element for dark matter annihilating into
        two charged pions and a netural pion through a scalar mediator.
        This function is used for RAMBO.

        Parameters
        ----------
        momenta : numpy.array
            List of momenta of the final state particles. The first momentum is
            for the positively charged pion, the second is for the negatively
            charged pion and the third is for the neutral pion.
        self : PseudoScalarMediatorParameters or PseudoScalarMediator object
            Parameter object for the pseudo-scalar mediator theory.

        Returns
        -------
        msqrd : float
            The squared matrix element for a given set of momenta.
        """
        p1 = momenta[0]
        p2 = momenta[1]
        p3 = momenta[2]
        P = sum(momenta)

        mx = self.mx
        mp = self.mp
        gpxx = self.gpxx
        gpuu = self.gpuu
        gpdd = self.gpdd
        gpGG = self.gpGG
        beta = self.beta
        widthp = self.width_p

        pxmag = np.sqrt(P[0] ** 2 / 4.0 - mx ** 2)

        px = np.array([P[0] / 2.0, 0.0, 0.0, pxmag])
        pxbar = np.array([P[0] / 2.0, 0.0, 0.0, -pxmag])

        """ Order beta^0 """
        beta0 = (
            gpxx ** 2
            * (
                -(b0 * fpi * gpGG * mdq)
                + b0 * fpi * gpGG * muq
                - b0 * fpi * gpdd * vh
                + b0 * fpi * gpuu * vh
            )
            ** 2
            * (mx ** 2 + MDot(px, pxbar))
        ) / (
            9.0
            * fpi ** 4
            * vh ** 2
            * (
                (-mp ** 2 + 2 * mpi ** 2 + mpi0 ** 2) ** 2
                + mp ** 2 * widthp ** 2
                - 4
                * (
                    mp ** 2
                    - 2 * mpi ** 2
                    - mpi0 ** 2
                    - MDot(p1, p2)
                    - MDot(p1, p3)
                    - MDot(p2, p3)
                )
                * (MDot(p1, p2) + MDot(p1, p3) + MDot(p2, p3))
            )
        )

        """ Order beta^1 """
        beta1 = (
            2
            * beta
            * gpxx ** 2
            * (
                -(b0 * fpi * gpGG * mdq)
                + b0 * fpi * gpGG * muq
                - b0 * fpi * gpdd * vh
                + b0 * fpi * gpuu * vh
            )
            * (
                b0 * mdq * vh
                + 2 * mpi ** 2 * vh
                - 2 * mpi0 ** 2 * vh
                + b0 * muq * vh
                + 4 * vh * MDot(p1, p2)
                - 2 * vh * MDot(p1, p3)
                - 2 * vh * MDot(p2, p3)
            )
            * (mx ** 2 + MDot(px, pxbar))
        ) / (
            9.0
            * fpi ** 4
            * vh ** 2
            * (
                (-mp ** 2 + 2 * mpi ** 2 + mpi0 ** 2) ** 2
                + mp ** 2 * widthp ** 2
                - 4
                * (
                    mp ** 2
                    - 2 * mpi ** 2
                    - mpi0 ** 2
                    - MDot(p1, p2)
                    - MDot(p1, p3)
                    - MDot(p2, p3)
                )
                * (MDot(p1, p2) + MDot(p1, p3) + MDot(p2, p3))
            )
        )

        """ Order beta^2 """

        beta2 = (
            beta ** 2
            * gpxx ** 2
            * (
                2
                * (
                    2 * b0 * fpi * gpGG * mdq
                    - 2 * b0 * fpi * gpGG * muq
                    + 2 * b0 * fpi * gpdd * vh
                    - 2 * b0 * fpi * gpuu * vh
                )
                * (
                    -(b0 * fpi * gpGG * mdq)
                    + b0 * fpi * gpGG * muq
                    - b0 * fpi * gpdd * vh
                    + b0 * fpi * gpuu * vh
                )
                - (
                    -(b0 * fpi * gpGG * mdq)
                    + b0 * fpi * gpGG * muq
                    - b0 * fpi * gpdd * vh
                    + b0 * fpi * gpuu * vh
                )
                ** 2
                + (
                    b0 * mdq * vh
                    + 2 * mpi ** 2 * vh
                    - 2 * mpi0 ** 2 * vh
                    + b0 * muq * vh
                    + 4 * vh * MDot(p1, p2)
                    - 2 * vh * MDot(p1, p3)
                    - 2 * vh * MDot(p2, p3)
                )
                ** 2
            )
            * (mx ** 2 + MDot(px, pxbar))
        ) / (
            9.0
            * fpi ** 4
            * vh ** 2
            * (
                (-mp ** 2 + 2 * mpi ** 2 + mpi0 ** 2) ** 2
                + mp ** 2 * widthp ** 2
                - 4
                * (
                    mp ** 2
                    - 2 * mpi ** 2
                    - mpi0 ** 2
                    - MDot(p1, p2)
                    - MDot(p1, p3)
                    - MDot(p2, p3)
                )
                * (MDot(p1, p2) + MDot(p1, p3) + MDot(p2, p3))
            )
        )

        msqrd = beta0 + beta1 + beta2

        return msqrd

    def msqrd_xx_to_p_to_pm0g(self, momenta):
        """
        Returns the squared matrix element for dark matter annihilating into
        two charged pions, a netural pion and a photon through a scalar
        mediator. This function is used for RAMBO.

        Parameters
        ----------
        momenta : numpy.array
            List of momenta of the final state particles. The first momentum is
            for the positively charged pion, the second is for the negatively
            charged pion, the third is for the neutral pion and the last is
            for the photon.
        self : PseudoScalarMediatorParameters or PseudoScalarMediator object
            Parameter object for the pseudo-scalar mediator theory.

        Returns
        -------
        msqrd : float
            The squared matrix element for a given set of momenta.
        """
        mx = self.mx
        mp = self.mp
        gpxx = self.gpxx
        gpuu = self.gpuu
        gpdd = self.gpdd
        gpGG = self.gpGG
        beta = self.beta
        widthp = self.width_p

        p1 = momenta[0]
        p2 = momenta[1]
        p3 = momenta[2]
        k = momenta[3]

        P = sum(momenta)

        pxmag = np.sqrt(P[0] ** 2 / 4.0 - mx ** 2)

        px = np.array([P[0] / 2.0, 0.0, 0.0, pxmag])
        pxbar = np.array([P[0] / 2.0, 0.0, 0.0, -pxmag])

        beta0 = -(
            gpxx ** 2
            * qe ** 2
            * (
                -(b0 * fpi * gpGG * mdq)
                + b0 * fpi * gpGG * muq
                - b0 * fpi * gpdd * vh
                + b0 * fpi * gpuu * vh
            )
            ** 2
            * (
                mpi ** 2 * MDot(k, p1) ** 2
                + mpi ** 2 * MDot(k, p2) ** 2
                - 2 * MDot(k, p1) * MDot(k, p2) * MDot(p1, p2)
            )
            * (mx ** 2 + MDot(px, pxbar))
        ) / (
            9.0
            * fpi ** 4
            * vh ** 2
            * MDot(k, p1) ** 2
            * MDot(k, p2) ** 2
            * (
                (-mp ** 2 + 2 * mpi ** 2 + mpi0 ** 2) ** 2
                + mp ** 2 * widthp ** 2
                - 4
                * (
                    mp ** 2
                    - 2 * mpi ** 2
                    - mpi0 ** 2
                    - MDot(k, p1)
                    - MDot(k, p2)
                    - MDot(k, p3)
                    - MDot(p1, p2)
                    - MDot(p1, p3)
                    - MDot(p2, p3)
                )
                * (
                    MDot(k, p1)
                    + MDot(k, p2)
                    + MDot(k, p3)
                    + MDot(p1, p2)
                    + MDot(p1, p3)
                    + MDot(p2, p3)
                )
            )
        )

        beta1 = (
            -2
            * beta
            * gpxx ** 2
            * qe ** 2
            * (
                -(b0 * fpi * gpGG * mdq)
                + b0 * fpi * gpGG * muq
                - b0 * fpi * gpdd * vh
                + b0 * fpi * gpuu * vh
            )
            * (
                mpi ** 2 * MDot(k, p1) ** 2
                + mpi ** 2 * MDot(k, p2) ** 2
                - 2 * MDot(k, p1) * MDot(k, p2) * MDot(p1, p2)
            )
            * (
                b0 * mdq * vh
                + 2 * mpi ** 2 * vh
                - 2 * mpi0 ** 2 * vh
                + b0 * muq * vh
                + 4 * vh * MDot(k, p1)
                + 4 * vh * MDot(k, p2)
                - 2 * vh * MDot(k, p3)
                + 4 * vh * MDot(p1, p2)
                - 2 * vh * MDot(p1, p3)
                - 2 * vh * MDot(p2, p3)
            )
            * (mx ** 2 + MDot(px, pxbar))
        ) / (
            9.0
            * fpi ** 4
            * vh ** 2
            * MDot(k, p1) ** 2
            * MDot(k, p2) ** 2
            * (
                (-mp ** 2 + 2 * mpi ** 2 + mpi0 ** 2) ** 2
                + mp ** 2 * widthp ** 2
                - 4
                * (
                    mp ** 2
                    - 2 * mpi ** 2
                    - mpi0 ** 2
                    - MDot(k, p1)
                    - MDot(k, p2)
                    - MDot(k, p3)
                    - MDot(p1, p2)
                    - MDot(p1, p3)
                    - MDot(p2, p3)
                )
                * (
                    MDot(k, p1)
                    + MDot(k, p2)
                    + MDot(k, p3)
                    + MDot(p1, p2)
                    + MDot(p1, p3)
                    + MDot(p2, p3)
                )
            )
        )

        beta2 = (
            beta ** 2
            * gpxx ** 2
            * qe ** 2
            * (
                mpi ** 2 * MDot(k, p1) ** 2
                + mpi ** 2 * MDot(k, p2) ** 2
                - 2 * MDot(k, p1) * MDot(k, p2) * MDot(p1, p2)
            )
            * (
                -2
                * (
                    2 * b0 * fpi * gpGG * mdq
                    - 2 * b0 * fpi * gpGG * muq
                    + 2 * b0 * fpi * gpdd * vh
                    - 2 * b0 * fpi * gpuu * vh
                )
                * (
                    -(b0 * fpi * gpGG * mdq)
                    + b0 * fpi * gpGG * muq
                    - b0 * fpi * gpdd * vh
                    + b0 * fpi * gpuu * vh
                )
                + (
                    -(b0 * fpi * gpGG * mdq)
                    + b0 * fpi * gpGG * muq
                    - b0 * fpi * gpdd * vh
                    + b0 * fpi * gpuu * vh
                )
                ** 2
                - (
                    b0 * mdq * vh
                    + 2 * mpi ** 2 * vh
                    - 2 * mpi0 ** 2 * vh
                    + b0 * muq * vh
                    + 4 * vh * MDot(k, p1)
                    + 4 * vh * MDot(k, p2)
                    - 2 * vh * MDot(k, p3)
                    + 4 * vh * MDot(p1, p2)
                    - 2 * vh * MDot(p1, p3)
                    - 2 * vh * MDot(p2, p3)
                )
                ** 2
            )
            * (mx ** 2 + MDot(px, pxbar))
        ) / (
            9.0
            * fpi ** 4
            * vh ** 2
            * MDot(k, p1) ** 2
            * MDot(k, p2) ** 2
            * (
                (-mp ** 2 + 2 * mpi ** 2 + mpi0 ** 2) ** 2
                + mp ** 2 * widthp ** 2
                - 4
                * (
                    mp ** 2
                    - 2 * mpi ** 2
                    - mpi0 ** 2
                    - MDot(k, p1)
                    - MDot(k, p2)
                    - MDot(k, p3)
                    - MDot(p1, p2)
                    - MDot(p1, p3)
                    - MDot(p2, p3)
                )
                * (
                    MDot(k, p1)
                    + MDot(k, p2)
                    + MDot(k, p3)
                    + MDot(p1, p2)
                    + MDot(p1, p3)
                    + MDot(p2, p3)
                )
            )
        )

        return beta0 + beta1 + beta2

    def msqrd_xx_to_p_to_000(self, momenta):
        """
        Returns the squared matrix element for dark matter annihilating into
        three neutral pions through a scalar mediator. This function is used
        for RAMBO.

        Parameters
        ----------
        momenta : numpy.array
            List of momenta of the final state particles, the three netural
            pions.
        self : PseudoScalarMediatorParameters or PseudoScalarMediator object
            Parameter object for the pseudo-scalar mediator theory.

        Returns
        -------
        msqrd : float
            The squared matrix element for a given set of momenta.
        """
        mx = self.mx
        mp = self.mp
        gpxx = self.gpxx
        gpuu = self.gpuu
        gpdd = self.gpdd
        gpGG = self.gpGG
        beta = self.beta
        widthp = self.width_p

        p1 = momenta[0]
        p2 = momenta[1]
        p3 = momenta[2]

        P = sum(momenta)

        pxmag = np.sqrt(P[0] ** 2 / 4.0 - mx ** 2)

        px = np.array([P[0] / 2.0, 0.0, 0.0, pxmag])
        pxbar = np.array([P[0] / 2.0, 0.0, 0.0, -pxmag])

        beta0 = (
            b0 ** 2
            * gpxx ** 2
            * (fpi * gpGG * mdq - fpi * gpGG * muq + fpi * gpdd * vh - fpi * gpuu * vh)
            ** 2
            * (mx ** 2 + MDot(px, pxbar))
        ) / (
            fpi ** 4
            * vh ** 2
            * (
                (mp ** 2 - 3 * mpi0 ** 2) ** 2
                + mp ** 2 * widthp ** 2
                - 4
                * (mp ** 2 - 3 * mpi0 ** 2 - MDot(p1, p2) - MDot(p1, p3) - MDot(p2, p3))
                * (MDot(p1, p2) + MDot(p1, p3) + MDot(p2, p3))
            )
        )

        beta1 = (
            -2
            * b0 ** 2
            * beta
            * gpxx ** 2
            * (fpi * gpGG * mdq - fpi * gpGG * muq + fpi * gpdd * vh - fpi * gpuu * vh)
            * (mdq * vh + muq * vh)
            * (mx ** 2 + MDot(px, pxbar))
        ) / (
            fpi ** 4
            * vh ** 2
            * (
                (mp ** 2 - 3 * mpi0 ** 2) ** 2
                + mp ** 2 * widthp ** 2
                - 4
                * (mp ** 2 - 3 * mpi0 ** 2 - MDot(p1, p2) - MDot(p1, p3) - MDot(p2, p3))
                * (MDot(p1, p2) + MDot(p1, p3) + MDot(p2, p3))
            )
        )

        beta2 = (
            b0 ** 2
            * beta ** 2
            * gpxx ** 2
            * (
                -3
                * (
                    fpi * gpGG * mdq
                    - fpi * gpGG * muq
                    + fpi * gpdd * vh
                    - fpi * gpuu * vh
                )
                ** 2
                + 2
                * (
                    fpi * gpGG * mdq
                    - fpi * gpGG * muq
                    + fpi * gpdd * vh
                    - fpi * gpuu * vh
                )
                * (
                    -4 * fpi * gpGG * mdq
                    + 4 * fpi * gpGG * muq
                    - 4 * fpi * gpdd * vh
                    + 4 * fpi * gpuu * vh
                )
                + (-(mdq * vh) - muq * vh) ** 2
            )
            * (mx ** 2 + MDot(px, pxbar))
        ) / (
            fpi ** 4
            * vh ** 2
            * (
                (mp ** 2 - 3 * mpi0 ** 2) ** 2
                + mp ** 2 * widthp ** 2
                - 4
                * (mp ** 2 - 3 * mpi0 ** 2 - MDot(p1, p2) - MDot(p1, p3) - MDot(p2, p3))
                * (MDot(p1, p2) + MDot(p1, p3) + MDot(p2, p3))
            )
        )

        msqrd = beta0 + beta1 + beta2

        return msqrd
